/*
** 2000-05-29
**
** The author disclaims copyright to this source code.  In place of
** a legal notice, here is a blessing:
**
**    May you do good and not evil.
**    May you find forgiveness for yourself and forgive others.
**    May you share freely, never taking more than you give.
**
*************************************************************************
** Driver template for the LEMON parser generator.
**
** The "lemon" program processes an LALR(1) input grammar file, then uses
** this template to construct a parser.  The "lemon" program inserts text
** at each "%%" line.  Also, any "P-a-r-s-e" identifer prefix (without the
** interstitial "-" characters) contained in this template is changed into
** the value of the %name directive from the grammar.  Otherwise, the content
** of this template is copied straight through into the generate parser
** source file.
**
** The following is the concatenation of all %include directives from the
** input grammar file:
*/
#include <stdio.h>
/************ Begin %include sections from the grammar ************************/
#line 1 "./mate_grammar.lemon"


/* mate_grammar.lemon
* MATE's configuration language grammar
*
* Copyright 2005, Luis E. Garcia Ontanon <luis@ontanon.org>
*
* Wireshark - Network traffic analyzer
* By Gerald Combs <gerald@wireshark.org>
* Copyright 1998 Gerald Combs
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

/*
 * XXX - there's a Lemon bug where this grammar produces a parser that
 * fails assertions; to work around it, we disable assert() failures.
 */
#ifndef NDEBUG
#define NDEBUG
#endif

#include "mate.h"
#include "mate_grammar.h"
#include <wsutil/file_util.h>

#define DUMMY void*

typedef struct _extraction {
	gchar* as;
	header_field_info* hfi;
	struct _extraction* next;
	struct _extraction* last;
} extraction_t;

typedef struct _pdu_criteria_t {
	AVPL* criterium_avpl;
	avpl_match_mode criterium_match_mode;
	accept_mode_t criterium_accept_mode;
} pdu_criteria_t;

typedef struct _gop_options {
	gop_tree_mode_t pdu_tree_mode;
	gboolean drop_unassigned;
	gboolean show_times;
	float expiration;
	float idle_timeout;
	float lifetime;
	AVPL* start;
	AVPL* stop;
	AVPL* extras;
} gop_options_t;

typedef struct _gog_statements {
	float expiration;
	gop_tree_mode_t gop_tree_mode;
	GPtrArray* transform_list;
	AVPL* extras;
	LoAL* current_gogkeys;
} gog_statement_t;

typedef struct _transf_match_t {
	avpl_match_mode match_mode;
	AVPL* avpl;
} transf_match_t;

typedef struct _transf_action_t {
	avpl_replace_mode replace_mode;
	AVPL* avpl;
} transf_action_t;

static void configuration_error(mate_config* mc, const gchar* fmt, ...) {
	static gchar error_buffer[256];
	const gchar* incl;
	gint i;
	mate_config_frame* current_frame;
	va_list list;

	va_start( list, fmt );
	g_vsnprintf(error_buffer,sizeof(error_buffer),fmt,list);
	va_end( list );

	i = (gint) mc->config_stack->len;

	while (i--) {

		if (i>0) {
			incl = "\n   included from: ";
		} else {
			incl = " ";
		}

		current_frame = (mate_config_frame *)g_ptr_array_index(mc->config_stack,(guint)i);

		g_string_append_printf(mc->config_error,"%s%s at line %u",incl, current_frame->filename, current_frame->linenum);
	}

	g_string_append_printf(mc->config_error,": %s\n",error_buffer);

	THROW(MateConfigError);

}

static AVPL_Transf* new_transform_elem(AVPL* match, AVPL* replace, avpl_match_mode match_mode, avpl_replace_mode replace_mode) {
	 AVPL_Transf* t = (AVPL_Transf *)g_malloc(sizeof(AVPL_Transf));

	 t->name = NULL;
	 t->match = match;
	 t->replace = replace;
	 t->match_mode = match_mode;
	 t->replace_mode = replace_mode;

	 t->map = NULL;
	 t->next = NULL;

	 return t;
}

static gchar* recolonize(mate_config* mc, gchar* s) {
	GString* str = g_string_new("");
	gchar** vec;
	gchar* r;
	guint i,v;
	gchar c;

	vec = g_strsplit(s,":",0);

	for (i = 0; vec[i]; i++) {
		g_ascii_strdown(vec[i], -1);

		v = 0;
		switch ( strlen(vec[i]) ) {
		 case 2:
			c = vec[i][1];
			vec[i][1] = vec[i][0];
			vec[i][0] = c;
			if (vec[i][0] >= '0' && vec[i][0] <= '9') {
				v += (vec[i][1] - '0' )*16;
			} else {
				v += (vec[i][1] - 'a' + 10)*16;
			}
		/* FALL THROUGH */
		 case 1:
			if (vec[i][0] >= '0' && vec[i][0] <= '9') {
				v += (vec[i][0] - '0' );
			} else {
				v += (vec[i][0] - 'a' + 10);
			}
		 case 0:
			break;
		 default:
			configuration_error(mc,"bad token %s",s);
		}

		g_string_append_printf(str,":%.2X",v);
	}

	g_strfreev(vec);

	g_string_erase(str,0,1);

	r = str->str;

	g_string_free(str,FALSE);

	return r;
}

#line 209 "mate_grammar.c"
/**************** End of %include directives **********************************/
/* These constants specify the various numeric values for terminal symbols
** in a format understandable to "makeheaders".  This section is blank unless
** "lemon" is run with the "-m" command-line option.
***************** Begin makeheaders token definitions *************************/
/**************** End makeheaders token definitions ***************************/

/* The next sections is a series of control #defines.
** various aspects of the generated parser.
**    YYCODETYPE         is the data type used to store the integer codes
**                       that represent terminal and non-terminal symbols.
**                       "unsigned char" is used if there are fewer than
**                       256 symbols.  Larger types otherwise.
**    YYNOCODE           is a number of type YYCODETYPE that is not used for
**                       any terminal or nonterminal symbol.
**    YYFALLBACK         If defined, this indicates that one or more tokens
**                       (also known as: "terminal symbols") have fall-back
**                       values which should be used if the original symbol
**                       would not parse.  This permits keywords to sometimes
**                       be used as identifiers, for example.
**    YYACTIONTYPE       is the data type used for "action codes" - numbers
**                       that indicate what to do in response to the next
**                       token.
**    MateParserTOKENTYPE     is the data type used for minor type for terminal
**                       symbols.  Background: A "minor type" is a semantic
**                       value associated with a terminal or non-terminal
**                       symbols.  For example, for an "ID" terminal symbol,
**                       the minor type might be the name of the identifier.
**                       Each non-terminal can have a different minor type.
**                       Terminal symbols all have the same minor type, though.
**                       This macros defines the minor type for terminal
**                       symbols.
**    YYMINORTYPE        is the data type used for all minor types.
**                       This is typically a union of many types, one of
**                       which is MateParserTOKENTYPE.  The entry in the union
**                       for terminal symbols is called "yy0".
**    YYSTACKDEPTH       is the maximum depth of the parser's stack.  If
**                       zero the stack is dynamically sized using realloc()
**    MateParserARG_SDECL     A static variable declaration for the %extra_argument
**    MateParserARG_PDECL     A parameter declaration for the %extra_argument
**    MateParserARG_STORE     Code to store %extra_argument into yypParser
**    MateParserARG_FETCH     Code to extract %extra_argument from yypParser
**    YYERRORSYMBOL      is the code number of the error symbol.  If not
**                       defined, then do no error processing.
**    YYNSTATE           the combined number of states.
**    YYNRULE            the number of rules in the grammar
**    YY_MAX_SHIFT       Maximum value for shift actions
**    YY_MIN_SHIFTREDUCE Minimum value for shift-reduce actions
**    YY_MAX_SHIFTREDUCE Maximum value for shift-reduce actions
**    YY_MIN_REDUCE      Maximum value for reduce actions
**    YY_ERROR_ACTION    The yy_action[] code for syntax error
**    YY_ACCEPT_ACTION   The yy_action[] code for accept
**    YY_NO_ACTION       The yy_action[] code for no-op
*/
#ifndef INTERFACE
# define INTERFACE 1
#endif
/************* Begin control #defines *****************************************/
#define YYCODETYPE unsigned char
#define YYNOCODE 139
#define YYACTIONTYPE unsigned short int
#define MateParserTOKENTYPE  gchar* 
typedef union {
  int yyinit;
  MateParserTOKENTYPE yy0;
  transf_match_t* yy16;
  pdu_criteria_t* yy41;
  AVPL_Transf* yy71;
  transf_action_t* yy77;
  float yy83;
  gchar* yy88;
  gop_tree_mode_t yy93;
  AVP* yy110;
  AVPL* yy170;
  gog_statement_t* yy171;
  avpl_match_mode yy172;
  accept_mode_t yy188;
  header_field_info* yy202;
  gop_options_t* yy203;
  avpl_replace_mode yy215;
  gboolean yy216;
  GPtrArray* yy231;
  LoAL* yy243;
  extraction_t* yy259;
} YYMINORTYPE;
#ifndef YYSTACKDEPTH
#define YYSTACKDEPTH 100
#endif
#define MateParserARG_SDECL  mate_config* mc ;
#define MateParserARG_PDECL , mate_config* mc 
#define MateParserARG_FETCH  mate_config* mc  = yypParser->mc 
#define MateParserARG_STORE yypParser->mc  = mc 
#define YYNSTATE             183
#define YYNRULE              149
#define YY_MAX_SHIFT         182
#define YY_MIN_SHIFTREDUCE   287
#define YY_MAX_SHIFTREDUCE   435
#define YY_MIN_REDUCE        436
#define YY_MAX_REDUCE        584
#define YY_ERROR_ACTION      585
#define YY_ACCEPT_ACTION     586
#define YY_NO_ACTION         587
/************* End control #defines *******************************************/

/* Define the yytestcase() macro to be a no-op if is not already defined
** otherwise.
**
** Applications can choose to define yytestcase() in the %include section
** to a macro that can assist in verifying code coverage.  For production
** code the yytestcase() macro should be turned off.  But it is useful
** for testing.
*/
#ifndef yytestcase
# define yytestcase(X)
#endif


/* Next are the tables used to determine what action to take based on the
** current state and lookahead token.  These tables are used to implement
** functions that take a state number and lookahead value and return an
** action integer.
**
** Suppose the action integer is N.  Then the action is determined as
** follows
**
**   0 <= N <= YY_MAX_SHIFT             Shift N.  That is, push the lookahead
**                                      token onto the stack and goto state N.
**
**   N between YY_MIN_SHIFTREDUCE       Shift to an arbitrary state then
**     and YY_MAX_SHIFTREDUCE           reduce by rule N-YY_MIN_SHIFTREDUCE.
**
**   N between YY_MIN_REDUCE            Reduce by rule N-YY_MIN_REDUCE
**     and YY_MAX_REDUCE
**
**   N == YY_ERROR_ACTION               A syntax error has occurred.
**
**   N == YY_ACCEPT_ACTION              The parser accepts its input.
**
**   N == YY_NO_ACTION                  No such action.  Denotes unused
**                                      slots in the yy_action[] table.
**
** The action table is constructed as a single large table named yy_action[].
** Given state S and lookahead X, the action is computed as either:
**
**    (A)   N = yy_action[ yy_shift_ofst[S] + X ]
**    (B)   N = yy_default[S]
**
** The (A) formula is preferred.  The B formula is used instead if:
**    (1)  The yy_shift_ofst[S]+X value is out of range, or
**    (2)  yy_lookahead[yy_shift_ofst[S]+X] is not equal to X, or
**    (3)  yy_shift_ofst[S] equal YY_SHIFT_USE_DFLT.
** (Implementation note: YY_SHIFT_USE_DFLT is chosen so that
** YY_SHIFT_USE_DFLT+X will be out of range for all possible lookaheads X.
** Hence only tests (1) and (2) need to be evaluated.)
**
** The formulas above are for computing the action when the lookahead is
** a terminal symbol.  If the lookahead is a non-terminal (as occurs after
** a reduce action) then the yy_reduce_ofst[] array is used in place of
** the yy_shift_ofst[] array and YY_REDUCE_USE_DFLT is used in place of
** YY_SHIFT_USE_DFLT.
**
** The following are the tables generated in this section:
**
**  yy_action[]        A single table containing all actions.
**  yy_lookahead[]     A table containing the lookahead for each entry in
**                     yy_action.  Used to detect hash collisions.
**  yy_shift_ofst[]    For each state, the offset into yy_action for
**                     shifting terminals.
**  yy_reduce_ofst[]   For each state, the offset into yy_action for
**                     shifting non-terminals after a reduce.
**  yy_default[]       Default action for each state.
**
*********** Begin parsing tables **********************************************/
#define YY_ACTTAB_COUNT (306)
static const YYACTIONTYPE yy_action[] = {
 /*     0 */   550,  182,  402,  181,  395,  396,   20,  398,    8,  308,
 /*    10 */     4,  138,  126,  111,  165,  358,  359,  360,  361,  318,
 /*    20 */   319,  374,  586,    1,  141,  372,  373,  315,  316,  317,
 /*    30 */   394,  101,  139,   47,   75,   43,   50,   11,  102,  103,
 /*    40 */   389,  306,  329,  366,  375,   71,  469,  397,    5,  376,
 /*    50 */   129,  402,  402,  402,  402,  402,  402,  399,  400,   70,
 /*    60 */   323,   20,  307,    4,   98,  339,  384,  330,    6,  362,
 /*    70 */   363,   99,  100,  339,  327,  328,  381,  392,   74,    3,
 /*    80 */   386,   63,   43,  365,  167,  166,  340,  387,  385,  129,
 /*    90 */   119,  393,  390,  161,  107,   66,   73,  106,   63,  157,
 /*   100 */   115,   19,    7,  116,  117,  378,   27,  105,   18,  118,
 /*   110 */   120,  109,  131,  132,   22,  133,   42,  142,  143,  144,
 /*   120 */   148,  149,  150,   11,   16,  112,  151,  156,  152,   64,
 /*   130 */    26,   14,  158,   15,   23,   25,   24,   50,   49,  364,
 /*   140 */    48,   31,   44,  124,   30,   29,   28,  102,   45,   52,
 /*   150 */   122,   51,   10,   57,   46,   55,   53,   54,   58,    2,
 /*   160 */    59,   34,   69,   33,  113,  114,   56,  135,   67,   68,
 /*   170 */   383,  140,   32,   39,  123,   17,   38,   35,   37,   21,
 /*   180 */    36,  162,  338,   40,  127,   41,  155,  130,  147,   12,
 /*   190 */   172,  178,  175,  136,  169,  104,  377,  310,  370,  313,
 /*   200 */   388,  368,  367,  108,   65,    9,  344,   60,  346,  379,
 /*   210 */   110,  352,  354,  356,  342,  348,  350,  341,   61,   13,
 /*   220 */   331,  121,   62,   72,  325,  125,  309,   78,  146,  128,
 /*   230 */   332,   77,  334,  336,  321,  305,  168,  137,   81,  154,
 /*   240 */    82,   80,   76,   83,   93,   91,   86,   84,  160,  164,
 /*   250 */    90,   95,  180,   87,   94,   96,   88,  134,  293,  294,
 /*   260 */   295,  418,  171,  296,  170,  145,   79,  291,  297,  298,
 /*   270 */   438,  299,  301,  173,  423,  438,  292,  174,  153,   85,
 /*   280 */   411,  302,  303,  438,  304,  431,  179,  176,  300,  438,
 /*   290 */   159,   97,  438,   89,  410,   92,  417,  177,  288,  438,
 /*   300 */   438,  287,  163,  438,  289,  290,
};
static const YYCODETYPE yy_lookahead[] = {
 /*     0 */     0,    1,   63,    3,    7,    8,   25,   10,   65,   66,
 /*    10 */    67,   11,   12,   13,   14,   44,   45,   46,   47,   29,
 /*    20 */    30,   47,  110,  111,   24,   51,   52,   26,   27,   28,
 /*    30 */   108,  109,   68,   86,   70,   53,   54,   24,    8,  106,
 /*    40 */   107,    5,   81,   10,   97,    4,   56,   50,   96,   97,
 /*    50 */    37,  112,  113,  114,  115,  116,  117,   60,   61,   98,
 /*    60 */     2,   25,   66,   67,  102,  103,   99,   81,   82,   48,
 /*    70 */    49,  104,  102,  103,   35,   36,    2,    5,   55,    4,
 /*    80 */    57,   40,   53,   50,    7,    8,    8,   57,    8,   37,
 /*    90 */    74,  108,  108,   74,  100,   69,   69,   72,   40,  100,
 /*   100 */    72,   77,   56,   72,   72,    8,   22,  105,   23,   75,
 /*   110 */    75,   89,   75,   75,   18,   75,   71,   75,   75,   75,
 /*   120 */    72,   72,   72,   24,   21,  105,   75,   72,   75,   55,
 /*   130 */    16,   59,   75,   58,   18,   20,   19,   54,   43,    8,
 /*   140 */    42,   39,   95,   73,   17,   16,   34,    8,   94,   87,
 /*   150 */    89,   88,   33,   86,   98,   90,   92,   91,   85,    4,
 /*   160 */    84,   17,   76,   16,  105,  105,   98,   80,   79,   78,
 /*   170 */    99,   64,   15,   22,  105,   21,   16,   18,   20,   23,
 /*   180 */    19,   13,  103,   18,  103,   22,   12,  105,   11,  101,
 /*   190 */    11,   13,   12,  103,    9,    6,    2,  105,    2,  105,
 /*   200 */   107,    2,    2,    5,   38,    4,    2,    4,    2,    2,
 /*   210 */     8,    2,    2,    2,    2,    2,    2,    2,   25,   32,
 /*   220 */     2,    5,   41,   31,    2,    8,    2,  126,  128,    8,
 /*   230 */     2,  127,    2,    2,    2,    2,   10,    8,  132,  134,
 /*   240 */   131,  133,    8,  130,  121,  123,  136,  129,  137,  125,
 /*   250 */   124,  119,  122,  133,  120,  118,  135,    5,    2,    2,
 /*   260 */     2,    2,    9,    2,   10,    5,    4,    2,    2,    2,
 /*   270 */   138,    2,    2,   10,    2,  138,    2,    9,    5,    4,
 /*   280 */     2,    2,    2,  138,    2,    2,    5,   10,    2,  138,
 /*   290 */     5,    4,  138,    4,    2,    4,    2,    9,    2,  138,
 /*   300 */   138,    2,    5,  138,    2,    2,
};
#define YY_SHIFT_USE_DFLT (306)
#define YY_SHIFT_COUNT    (182)
#define YY_SHIFT_MIN      (-29)
#define YY_SHIFT_MAX      (303)
static const short yy_shift_ofst[] = {
 /*     0 */   306,    0,  -19,   -3,  -10,  -18,   13,   30,   36,   29,
 /*    10 */    78,   80,   52,   78,   -3,   -3,  -29,  -29,  -26,    1,
 /*    20 */     1,  -26,   33,   33,   33,   33,   21,   21,   39,   21,
 /*    30 */    21,   21,   21,   21,   21,   33,   33,   33,   21,   21,
 /*    40 */    33,   21,   46,   97,   84,   85,   96,   99,   46,   46,
 /*    50 */    46,   84,  103,  114,  115,  117,  116,   99,   83,   95,
 /*    60 */    98,   46,  131,   78,   80,   78,   46,  102,  127,  129,
 /*    70 */   112,  119,   78,   46,  139,   46,  155,  144,  147,  157,
 /*    80 */   151,  154,  160,  158,  161,  159,  163,  156,  154,  165,
 /*    90 */   168,  174,  177,  178,  180,  179,  185,  189,   58,   74,
 /*   100 */    41,   72,   75,   23,   77,  194,  196,  199,  200,  198,
 /*   110 */   201,  202,  204,  206,  207,  209,  210,  211,  212,  213,
 /*   120 */   214,  215,  216,  203,  193,  181,  217,  218,  166,  221,
 /*   130 */   222,  228,  230,  231,  232,  252,  187,  192,  229,  224,
 /*   140 */   233,  234,  256,  257,  258,  259,  260,  262,  261,  266,
 /*   150 */   267,  269,  270,  272,  273,  275,  279,  280,  282,  283,
 /*   160 */   285,  286,  289,  294,  297,  291,  296,  299,  302,  226,
 /*   170 */   303,  254,  253,  265,  263,  268,  274,  277,  288,  278,
 /*   180 */   281,  287,  292,
};
#define YY_REDUCE_USE_DFLT (-89)
#define YY_REDUCE_COUNT (97)
#define YY_REDUCE_MIN   (-88)
#define YY_REDUCE_MAX   (137)
static const short yy_reduce_ofst[] = {
 /*     0 */   -88,  -61,  -57,  -78,  -36,  -53,  -39,  -67,   -4,  -48,
 /*    10 */   -38,  -33,  -14,  -30,  -17,  -16,   16,   19,   -6,   26,
 /*    20 */    27,   -1,   25,   28,   31,   32,   34,   35,   24,   37,
 /*    30 */    38,   40,   42,   43,   44,   48,   49,   50,   51,   53,
 /*    40 */    55,   57,    2,   45,   22,   47,   54,   56,   20,   59,
 /*    50 */    60,   61,   63,   62,   64,   66,   65,   68,   67,   73,
 /*    60 */    76,   69,   70,   79,   71,   81,   82,   87,   89,   91,
 /*    70 */    86,   88,   90,   92,   93,   94,  107,  100,  104,  101,
 /*    80 */   105,  108,  106,  109,  113,  118,  111,  110,  120,  121,
 /*    90 */   124,  126,  122,  130,  123,  134,  132,  137,
};
static const YYACTIONTYPE yy_default[] = {
 /*     0 */   552,  585,  460,  585,  461,  529,  531,  585,  460,  585,
 /*    10 */   585,  585,  585,  585,  585,  585,  585,  585,  585,  463,
 /*    20 */   463,  585,  585,  585,  585,  585,  585,  585,  475,  585,
 /*    30 */   585,  585,  585,  585,  585,  585,  585,  585,  585,  585,
 /*    40 */   585,  585,  585,  585,  500,  518,  520,  531,  585,  585,
 /*    50 */   585,  500,  498,  492,  506,  504,  502,  531,  529,  496,
 /*    60 */   494,  585,  585,  585,  585,  585,  585,  486,  484,  482,
 /*    70 */   473,  471,  585,  585,  585,  585,  585,  571,  570,  569,
 /*    80 */   579,  578,  577,  576,  575,  574,  584,  583,  578,  582,
 /*    90 */   581,  573,  568,  565,  564,  563,  562,  561,  585,  585,
 /*   100 */   585,  585,  540,  585,  585,  585,  585,  585,  585,  585,
 /*   110 */   585,  585,  585,  585,  585,  585,  585,  585,  585,  585,
 /*   120 */   585,  585,  585,  585,  585,  585,  585,  585,  585,  585,
 /*   130 */   585,  585,  585,  585,  585,  585,  585,  585,  585,  585,
 /*   140 */   585,  585,  585,  585,  585,  585,  585,  585,  585,  585,
 /*   150 */   585,  585,  585,  585,  585,  585,  585,  585,  585,  585,
 /*   160 */   585,  585,  585,  585,  585,  585,  585,  585,  585,  585,
 /*   170 */   585,  585,  585,  585,  585,  585,  585,  585,  585,  585,
 /*   180 */   585,  585,  585,
};
/********** End of lemon-generated parsing tables *****************************/

/* The next table maps tokens (terminal symbols) into fallback tokens.
** If a construct like the following:
**
**      %fallback ID X Y Z.
**
** appears in the grammar, then ID becomes a fallback token for X, Y,
** and Z.  Whenever one of the tokens X, Y, or Z is input to the parser
** but it does not parse, the type of the token is changed to ID and
** the parse is retried before an error is thrown.
**
** This feature can be used, for example, to cause some keywords in a language
** to revert to identifiers if they keyword does not apply in the context where
** it appears.
*/
#ifdef YYFALLBACK
static const YYCODETYPE yyFallback[] = {
};
#endif /* YYFALLBACK */

/* The following structure represents a single element of the
** parser's stack.  Information stored includes:
**
**   +  The state number for the parser at this level of the stack.
**
**   +  The value of the token stored at this level of the stack.
**      (In other words, the "major" token.)
**
**   +  The semantic value stored at this level of the stack.  This is
**      the information used by the action routines in the grammar.
**      It is sometimes called the "minor" token.
**
** After the "shift" half of a SHIFTREDUCE action, the stateno field
** actually contains the reduce action for the second half of the
** SHIFTREDUCE.
*/
struct yyStackEntry {
  YYACTIONTYPE stateno;  /* The state-number, or reduce action in SHIFTREDUCE */
  YYCODETYPE major;      /* The major token value.  This is the code
                         ** number for the token at this stack level */
  YYMINORTYPE minor;     /* The user-supplied minor token value.  This
                         ** is the value of the token  */
};
typedef struct yyStackEntry yyStackEntry;

/* The state of the parser is completely contained in an instance of
** the following structure */
struct yyParser {
  yyStackEntry *yytos;          /* Pointer to top element of the stack */
#ifdef YYTRACKMAXSTACKDEPTH
  int yyhwm;                    /* High-water mark of the stack */
#endif
#ifndef YYNOERRORRECOVERY
  int yyerrcnt;                 /* Shifts left before out of the error */
#endif
  MateParserARG_SDECL                /* A place to hold %extra_argument */
#if YYSTACKDEPTH<=0
  int yystksz;                  /* Current side of the stack */
  yyStackEntry *yystack;        /* The parser's stack */
  yyStackEntry yystk0;          /* First stack entry */
#else
  yyStackEntry yystack[YYSTACKDEPTH];  /* The parser's stack */
#endif
};
typedef struct yyParser yyParser;

#ifndef NDEBUG
#include <stdio.h>
static FILE *yyTraceFILE = 0;
static char *yyTracePrompt = 0;
#endif /* NDEBUG */

#ifndef NDEBUG
/*
** Turn parser tracing on by giving a stream to which to write the trace
** and a prompt to preface each trace message.  Tracing is turned off
** by making either argument NULL
**
** Inputs:
** <ul>
** <li> A FILE* to which trace output should be written.
**      If NULL, then tracing is turned off.
** <li> A prefix string written at the beginning of every
**      line of trace output.  If NULL, then tracing is
**      turned off.
** </ul>
**
** Outputs:
** None.
*/
void MateParserTrace(FILE *TraceFILE, char *zTracePrompt){
  yyTraceFILE = TraceFILE;
  yyTracePrompt = zTracePrompt;
  if( yyTraceFILE==0 ) yyTracePrompt = 0;
  else if( yyTracePrompt==0 ) yyTraceFILE = 0;
}
#endif /* NDEBUG */

#ifndef NDEBUG
/* For tracing shifts, the names of all terminals and nonterminals
** are required.  The following table supplies these names */
static const char *const yyTokenName[] = {
  "$",             "DONE_KW",       "SEMICOLON",     "DEBUG_KW",    
  "OPEN_BRACE",    "CLOSE_BRACE",   "FILENAME_KW",   "QUOTED",      
  "NAME",          "LEVEL_KW",      "INTEGER",       "PDU_KW",      
  "GOP_KW",        "GOG_KW",        "DEFAULT_KW",    "LAST_EXTRACTED_KW",
  "DROP_UNASSIGNED_KW",  "DISCARD_PDU_DATA_KW",  "EXPIRATION_KW",  "IDLE_TIMEOUT_KW",
  "LIFETIME_KW",   "SHOW_TREE_KW",  "SHOW_TIMES_KW",  "GOP_TREE_KW", 
  "TRANSFORM_KW",  "MATCH_KW",      "STRICT_KW",     "EVERY_KW",    
  "LOOSE_KW",      "REPLACE_KW",    "INSERT_KW",     "PROTO_KW",    
  "TRANSPORT_KW",  "PAYLOAD_KW",    "CRITERIA_KW",   "ACCEPT_KW",   
  "REJECT_KW",     "EXTRACT_KW",    "FROM_KW",       "LAST_PDU_KW", 
  "SLASH",         "ON_KW",         "START_KW",      "STOP_KW",     
  "NO_TREE_KW",    "PDU_TREE_KW",   "FRAME_TREE_KW",  "BASIC_TREE_KW",
  "TRUE_KW",       "FALSE_KW",      "FLOATING",      "NULL_TREE_KW",
  "FULL_TREE_KW",  "MEMBER_KW",     "EXTRA_KW",      "COMMA",       
  "OPEN_PARENS",   "CLOSE_PARENS",  "AVP_OPERATOR",  "PIPE",        
  "DOTED_IP",      "COLONIZED",     "error",         "transform_decl",
  "transform_body",  "transform_statements",  "transform_statement",  "transform_match",
  "transform_action",  "match_mode",    "action_mode",   "gop_name",    
  "time_value",    "pdu_name",      "gop_tree_mode",  "true_false",  
  "criteria_statement",  "accept_mode",   "pdu_drop_unassigned_statement",  "discard_pdu_data_statement",
  "last_extracted_statement",  "extraction_statement",  "extraction_statements",  "gop_options", 
  "gop_start_statement",  "gop_stop_statement",  "extra_statement",  "gop_drop_unassigned_statement",
  "show_goptree_statement",  "show_times_statement",  "gop_expiration_statement",  "idle_timeout_statement",
  "lifetime_statement",  "gog_statements",  "gog_expiration_statement",  "gog_goptree_statement",
  "gog_key_statements",  "gog_key_statement",  "transform_list_statement",  "transform",   
  "gop_tree_type",  "payload_statement",  "proto_stack",   "field",       
  "transform_list",  "avpl",          "avps",          "avp",         
  "value",         "avp_oneoff",    "mate_config",   "decls",       
  "decl",          "pdu_decl",      "gop_decl",      "gog_decl",    
  "defaults_decl",  "debug_decl",    "dbgfile_default",  "dbglevel_default",
  "pdu_dbglevel_default",  "gop_dbglevel_default",  "gog_dbglevel_default",  "pdu_defaults",
  "gop_defaults",  "gog_defaults",  "pdu_last_extracted_default",  "pdu_drop_unassigned_default",
  "pdu_discard_default",  "gop_expiration_default",  "gop_idle_timeout_default",  "gop_lifetime_default",
  "gop_drop_unassigned_default",  "gop_tree_mode_default",  "gop_show_times_default",  "gog_expiration_default",
  "gog_goptree_default",  "gog_show_times_default",
};
#endif /* NDEBUG */

#ifndef NDEBUG
/* For tracing reduce actions, the names of all rules are required.
*/
static const char *const yyRuleName[] = {
 /*   0 */ "dbgfile_default ::= FILENAME_KW QUOTED SEMICOLON",
 /*   1 */ "dbgfile_default ::= FILENAME_KW NAME SEMICOLON",
 /*   2 */ "dbglevel_default ::= LEVEL_KW INTEGER SEMICOLON",
 /*   3 */ "pdu_dbglevel_default ::= PDU_KW LEVEL_KW INTEGER SEMICOLON",
 /*   4 */ "gop_dbglevel_default ::= GOP_KW LEVEL_KW INTEGER SEMICOLON",
 /*   5 */ "gog_dbglevel_default ::= GOG_KW LEVEL_KW INTEGER SEMICOLON",
 /*   6 */ "pdu_last_extracted_default ::= LAST_EXTRACTED_KW true_false SEMICOLON",
 /*   7 */ "pdu_drop_unassigned_default ::= DROP_UNASSIGNED_KW true_false SEMICOLON",
 /*   8 */ "pdu_discard_default ::= DISCARD_PDU_DATA_KW true_false SEMICOLON",
 /*   9 */ "gop_expiration_default ::= EXPIRATION_KW time_value SEMICOLON",
 /*  10 */ "gop_idle_timeout_default ::= IDLE_TIMEOUT_KW time_value SEMICOLON",
 /*  11 */ "gop_lifetime_default ::= LIFETIME_KW time_value SEMICOLON",
 /*  12 */ "gop_drop_unassigned_default ::= DROP_UNASSIGNED_KW true_false SEMICOLON",
 /*  13 */ "gop_tree_mode_default ::= SHOW_TREE_KW gop_tree_mode SEMICOLON",
 /*  14 */ "gop_show_times_default ::= SHOW_TIMES_KW true_false SEMICOLON",
 /*  15 */ "gog_expiration_default ::= EXPIRATION_KW time_value SEMICOLON",
 /*  16 */ "gog_goptree_default ::= GOP_TREE_KW gop_tree_type SEMICOLON",
 /*  17 */ "gog_show_times_default ::= SHOW_TIMES_KW true_false SEMICOLON",
 /*  18 */ "transform_decl ::= TRANSFORM_KW NAME transform_body SEMICOLON",
 /*  19 */ "transform_body ::= OPEN_BRACE transform_statements CLOSE_BRACE",
 /*  20 */ "transform_statements ::= transform_statements transform_statement",
 /*  21 */ "transform_statements ::= transform_statement",
 /*  22 */ "transform_statement ::= transform_match transform_action SEMICOLON",
 /*  23 */ "transform_match ::= MATCH_KW match_mode avpl",
 /*  24 */ "transform_match ::=",
 /*  25 */ "transform_action ::=",
 /*  26 */ "transform_action ::= action_mode avpl",
 /*  27 */ "match_mode ::=",
 /*  28 */ "match_mode ::= STRICT_KW",
 /*  29 */ "match_mode ::= EVERY_KW",
 /*  30 */ "match_mode ::= LOOSE_KW",
 /*  31 */ "action_mode ::= REPLACE_KW",
 /*  32 */ "action_mode ::= INSERT_KW",
 /*  33 */ "action_mode ::=",
 /*  34 */ "pdu_decl ::= PDU_KW NAME PROTO_KW field TRANSPORT_KW proto_stack OPEN_BRACE payload_statement extraction_statements transform_list_statement criteria_statement pdu_drop_unassigned_statement discard_pdu_data_statement last_extracted_statement CLOSE_BRACE SEMICOLON",
 /*  35 */ "payload_statement ::=",
 /*  36 */ "payload_statement ::= PAYLOAD_KW proto_stack SEMICOLON",
 /*  37 */ "criteria_statement ::=",
 /*  38 */ "criteria_statement ::= CRITERIA_KW accept_mode match_mode avpl SEMICOLON",
 /*  39 */ "accept_mode ::=",
 /*  40 */ "accept_mode ::= ACCEPT_KW",
 /*  41 */ "accept_mode ::= REJECT_KW",
 /*  42 */ "extraction_statements ::= extraction_statements extraction_statement",
 /*  43 */ "extraction_statements ::= extraction_statement",
 /*  44 */ "extraction_statement ::= EXTRACT_KW NAME FROM_KW field SEMICOLON",
 /*  45 */ "pdu_drop_unassigned_statement ::= DROP_UNASSIGNED_KW true_false SEMICOLON",
 /*  46 */ "pdu_drop_unassigned_statement ::=",
 /*  47 */ "discard_pdu_data_statement ::= DISCARD_PDU_DATA_KW true_false SEMICOLON",
 /*  48 */ "discard_pdu_data_statement ::=",
 /*  49 */ "last_extracted_statement ::= LAST_PDU_KW true_false SEMICOLON",
 /*  50 */ "last_extracted_statement ::=",
 /*  51 */ "proto_stack ::= proto_stack SLASH field",
 /*  52 */ "proto_stack ::= field",
 /*  53 */ "field ::= NAME",
 /*  54 */ "gop_decl ::= GOP_KW NAME ON_KW pdu_name MATCH_KW avpl OPEN_BRACE gop_start_statement gop_stop_statement extra_statement transform_list_statement gop_expiration_statement idle_timeout_statement lifetime_statement gop_drop_unassigned_statement show_goptree_statement show_times_statement CLOSE_BRACE SEMICOLON",
 /*  55 */ "gop_drop_unassigned_statement ::= DROP_UNASSIGNED_KW true_false SEMICOLON",
 /*  56 */ "gop_drop_unassigned_statement ::=",
 /*  57 */ "gop_start_statement ::= START_KW avpl SEMICOLON",
 /*  58 */ "gop_start_statement ::=",
 /*  59 */ "gop_stop_statement ::= STOP_KW avpl SEMICOLON",
 /*  60 */ "gop_stop_statement ::=",
 /*  61 */ "show_goptree_statement ::= SHOW_TREE_KW gop_tree_mode SEMICOLON",
 /*  62 */ "show_goptree_statement ::=",
 /*  63 */ "show_times_statement ::= SHOW_TIMES_KW true_false SEMICOLON",
 /*  64 */ "show_times_statement ::=",
 /*  65 */ "gop_expiration_statement ::= EXPIRATION_KW time_value SEMICOLON",
 /*  66 */ "gop_expiration_statement ::=",
 /*  67 */ "idle_timeout_statement ::= IDLE_TIMEOUT_KW time_value SEMICOLON",
 /*  68 */ "idle_timeout_statement ::=",
 /*  69 */ "lifetime_statement ::= LIFETIME_KW time_value SEMICOLON",
 /*  70 */ "lifetime_statement ::=",
 /*  71 */ "gop_tree_mode ::= NO_TREE_KW",
 /*  72 */ "gop_tree_mode ::= PDU_TREE_KW",
 /*  73 */ "gop_tree_mode ::= FRAME_TREE_KW",
 /*  74 */ "gop_tree_mode ::= BASIC_TREE_KW",
 /*  75 */ "true_false ::= TRUE_KW",
 /*  76 */ "true_false ::= FALSE_KW",
 /*  77 */ "pdu_name ::= NAME",
 /*  78 */ "time_value ::= FLOATING",
 /*  79 */ "time_value ::= INTEGER",
 /*  80 */ "gog_decl ::= GOG_KW NAME OPEN_BRACE gog_key_statements extra_statement transform_list_statement gog_expiration_statement gog_goptree_statement show_times_statement CLOSE_BRACE SEMICOLON",
 /*  81 */ "gog_goptree_statement ::= GOP_TREE_KW gop_tree_type SEMICOLON",
 /*  82 */ "gog_goptree_statement ::=",
 /*  83 */ "gog_expiration_statement ::= EXPIRATION_KW time_value SEMICOLON",
 /*  84 */ "gog_expiration_statement ::=",
 /*  85 */ "gop_tree_type ::= NULL_TREE_KW",
 /*  86 */ "gop_tree_type ::= FULL_TREE_KW",
 /*  87 */ "gop_tree_type ::= BASIC_TREE_KW",
 /*  88 */ "gog_key_statements ::= gog_key_statements gog_key_statement",
 /*  89 */ "gog_key_statements ::= gog_key_statement",
 /*  90 */ "gog_key_statement ::= MEMBER_KW gop_name avpl SEMICOLON",
 /*  91 */ "gop_name ::= NAME",
 /*  92 */ "extra_statement ::= EXTRA_KW avpl SEMICOLON",
 /*  93 */ "extra_statement ::=",
 /*  94 */ "transform_list_statement ::= TRANSFORM_KW transform_list SEMICOLON",
 /*  95 */ "transform_list_statement ::=",
 /*  96 */ "transform_list ::= transform_list COMMA transform",
 /*  97 */ "transform_list ::= transform",
 /*  98 */ "transform ::= NAME",
 /*  99 */ "avpl ::= OPEN_PARENS avps CLOSE_PARENS",
 /* 100 */ "avpl ::= OPEN_PARENS CLOSE_PARENS",
 /* 101 */ "avps ::= avps COMMA avp",
 /* 102 */ "avps ::= avp",
 /* 103 */ "avp ::= NAME AVP_OPERATOR value",
 /* 104 */ "avp ::= NAME",
 /* 105 */ "avp ::= NAME OPEN_BRACE avp_oneoff CLOSE_BRACE",
 /* 106 */ "avp_oneoff ::= avp_oneoff PIPE value",
 /* 107 */ "avp_oneoff ::= value",
 /* 108 */ "value ::= QUOTED",
 /* 109 */ "value ::= NAME",
 /* 110 */ "value ::= FLOATING",
 /* 111 */ "value ::= INTEGER",
 /* 112 */ "value ::= DOTED_IP",
 /* 113 */ "value ::= COLONIZED",
 /* 114 */ "mate_config ::= decls",
 /* 115 */ "decls ::= decls decl",
 /* 116 */ "decls ::=",
 /* 117 */ "decl ::= pdu_decl",
 /* 118 */ "decl ::= gop_decl",
 /* 119 */ "decl ::= gog_decl",
 /* 120 */ "decl ::= transform_decl",
 /* 121 */ "decl ::= defaults_decl",
 /* 122 */ "decl ::= debug_decl",
 /* 123 */ "decl ::= DONE_KW SEMICOLON",
 /* 124 */ "debug_decl ::= DEBUG_KW OPEN_BRACE dbgfile_default dbglevel_default pdu_dbglevel_default gop_dbglevel_default gog_dbglevel_default CLOSE_BRACE SEMICOLON",
 /* 125 */ "dbgfile_default ::=",
 /* 126 */ "dbglevel_default ::=",
 /* 127 */ "pdu_dbglevel_default ::=",
 /* 128 */ "gop_dbglevel_default ::=",
 /* 129 */ "gog_dbglevel_default ::=",
 /* 130 */ "defaults_decl ::= DEFAULT_KW OPEN_BRACE pdu_defaults gop_defaults gog_defaults CLOSE_BRACE SEMICOLON",
 /* 131 */ "pdu_defaults ::= PDU_KW OPEN_BRACE pdu_last_extracted_default pdu_drop_unassigned_default pdu_discard_default CLOSE_BRACE SEMICOLON",
 /* 132 */ "pdu_defaults ::=",
 /* 133 */ "pdu_last_extracted_default ::=",
 /* 134 */ "pdu_drop_unassigned_default ::=",
 /* 135 */ "pdu_discard_default ::=",
 /* 136 */ "gop_defaults ::= GOP_KW OPEN_BRACE gop_expiration_default gop_idle_timeout_default gop_lifetime_default gop_drop_unassigned_default gop_tree_mode_default gop_show_times_default CLOSE_BRACE SEMICOLON",
 /* 137 */ "gop_defaults ::=",
 /* 138 */ "gop_expiration_default ::=",
 /* 139 */ "gop_idle_timeout_default ::=",
 /* 140 */ "gop_lifetime_default ::=",
 /* 141 */ "gop_drop_unassigned_default ::=",
 /* 142 */ "gop_tree_mode_default ::=",
 /* 143 */ "gop_show_times_default ::=",
 /* 144 */ "gog_defaults ::= GOG_KW OPEN_BRACE gog_expiration_default gop_tree_mode_default gog_goptree_default gog_show_times_default CLOSE_BRACE SEMICOLON",
 /* 145 */ "gog_defaults ::=",
 /* 146 */ "gog_expiration_default ::=",
 /* 147 */ "gog_goptree_default ::=",
 /* 148 */ "gog_show_times_default ::=",
};
#endif /* NDEBUG */


#if YYSTACKDEPTH<=0
/*
** Try to increase the size of the parser stack.  Return the number
** of errors.  Return 0 on success.
*/
static int yyGrowStack(yyParser *p){
  int newSize;
  int idx;
  yyStackEntry *pNew;

  newSize = p->yystksz*2 + 100;
  idx = p->yytos ? (int)(p->yytos - p->yystack) : 0;
  if( p->yystack==&p->yystk0 ){
    pNew = malloc(newSize*sizeof(pNew[0]));
    if( pNew ) pNew[0] = p->yystk0;
  }else{
    pNew = realloc(p->yystack, newSize*sizeof(pNew[0]));
  }
  if( pNew ){
    p->yystack = pNew;
    p->yytos = &p->yystack[idx];
#ifndef NDEBUG
    if( yyTraceFILE ){
      fprintf(yyTraceFILE,"%sStack grows from %d to %d entries.\n",
              yyTracePrompt, p->yystksz, newSize);
    }
#endif
    p->yystksz = newSize;
  }
  return pNew==0;
}
#endif

/* Datatype of the argument to the memory allocated passed as the
** second argument to MateParserAlloc() below.  This can be changed by
** putting an appropriate #define in the %include section of the input
** grammar.
*/
#ifndef YYMALLOCARGTYPE
# define YYMALLOCARGTYPE size_t
#endif

/* Initialize a new parser that has already been allocated.
*/
static void MateParserInit(void *yypParser){
  yyParser *pParser = (yyParser*)yypParser;
#ifdef YYTRACKMAXSTACKDEPTH
  pParser->yyhwm = 0;
#endif
#if YYSTACKDEPTH<=0
  pParser->yytos = NULL;
  pParser->yystack = NULL;
  pParser->yystksz = 0;
  if( yyGrowStack(pParser) ){
    pParser->yystack = &pParser->yystk0;
    pParser->yystksz = 1;
  }
#endif
#ifndef YYNOERRORRECOVERY
  pParser->yyerrcnt = -1;
#endif
  pParser->yytos = pParser->yystack;
  pParser->yystack[0].stateno = 0;
  pParser->yystack[0].major = 0;
}

#ifndef MateParser_ENGINEALWAYSONSTACK
/*
** This function allocates a new parser.
** The only argument is a pointer to a function which works like
** malloc.
**
** Inputs:
** A pointer to the function used to allocate memory.
**
** Outputs:
** A pointer to a parser.  This pointer is used in subsequent calls
** to MateParser and MateParserFree.
*/
void *MateParserAlloc(void *(*mallocProc)(YYMALLOCARGTYPE)){
  yyParser *pParser;
  pParser = (yyParser*)(*mallocProc)( (YYMALLOCARGTYPE)sizeof(yyParser) );
  if( pParser ) MateParserInit(pParser);
  return pParser;
}
#endif /* MateParser_ENGINEALWAYSONSTACK */


/* The following function deletes the "minor type" or semantic value
** associated with a symbol.  The symbol can be either a terminal
** or nonterminal. "yymajor" is the symbol code, and "yypminor" is
** a pointer to the value to be deleted.  The code used to do the
** deletions is derived from the %destructor and/or %token_destructor
** directives of the input grammar.
*/
static void yy_destructor(
  yyParser *yypParser,    /* The parser */
  YYCODETYPE yymajor,     /* Type code for object to destroy */
  YYMINORTYPE *yypminor   /* The object to be destroyed */
){
  MateParserARG_FETCH;
  switch( yymajor ){
    /* Here is inserted the actions which take place when a
    ** terminal or non-terminal is destroyed.  This can happen
    ** when the symbol is popped from the stack during a
    ** reduce or during error processing or when a parser is
    ** being destroyed before it is finished parsing.
    **
    ** Note: during a reduce, the only symbols destroyed are those
    ** which appear on the RHS of the rule, but which are *not* used
    ** inside the C code.
    */
/********* Begin destructor definitions ***************************************/
      /* TERMINAL Destructor */
    case 1: /* DONE_KW */
    case 2: /* SEMICOLON */
    case 3: /* DEBUG_KW */
    case 4: /* OPEN_BRACE */
    case 5: /* CLOSE_BRACE */
    case 6: /* FILENAME_KW */
    case 7: /* QUOTED */
    case 8: /* NAME */
    case 9: /* LEVEL_KW */
    case 10: /* INTEGER */
    case 11: /* PDU_KW */
    case 12: /* GOP_KW */
    case 13: /* GOG_KW */
    case 14: /* DEFAULT_KW */
    case 15: /* LAST_EXTRACTED_KW */
    case 16: /* DROP_UNASSIGNED_KW */
    case 17: /* DISCARD_PDU_DATA_KW */
    case 18: /* EXPIRATION_KW */
    case 19: /* IDLE_TIMEOUT_KW */
    case 20: /* LIFETIME_KW */
    case 21: /* SHOW_TREE_KW */
    case 22: /* SHOW_TIMES_KW */
    case 23: /* GOP_TREE_KW */
    case 24: /* TRANSFORM_KW */
    case 25: /* MATCH_KW */
    case 26: /* STRICT_KW */
    case 27: /* EVERY_KW */
    case 28: /* LOOSE_KW */
    case 29: /* REPLACE_KW */
    case 30: /* INSERT_KW */
    case 31: /* PROTO_KW */
    case 32: /* TRANSPORT_KW */
    case 33: /* PAYLOAD_KW */
    case 34: /* CRITERIA_KW */
    case 35: /* ACCEPT_KW */
    case 36: /* REJECT_KW */
    case 37: /* EXTRACT_KW */
    case 38: /* FROM_KW */
    case 39: /* LAST_PDU_KW */
    case 40: /* SLASH */
    case 41: /* ON_KW */
    case 42: /* START_KW */
    case 43: /* STOP_KW */
    case 44: /* NO_TREE_KW */
    case 45: /* PDU_TREE_KW */
    case 46: /* FRAME_TREE_KW */
    case 47: /* BASIC_TREE_KW */
    case 48: /* TRUE_KW */
    case 49: /* FALSE_KW */
    case 50: /* FLOATING */
    case 51: /* NULL_TREE_KW */
    case 52: /* FULL_TREE_KW */
    case 53: /* MEMBER_KW */
    case 54: /* EXTRA_KW */
    case 55: /* COMMA */
    case 56: /* OPEN_PARENS */
    case 57: /* CLOSE_PARENS */
    case 58: /* AVP_OPERATOR */
    case 59: /* PIPE */
    case 60: /* DOTED_IP */
    case 61: /* COLONIZED */
{
#line 188 "./mate_grammar.lemon"

	(void) mc; /* Mark unused, similar to Q_UNUSED */
	if ((yypminor->yy0)) g_free((yypminor->yy0));

#line 991 "mate_grammar.c"
}
      break;
/********* End destructor definitions *****************************************/
    default:  break;   /* If no destructor action specified: do nothing */
  }
}

/*
** Pop the parser's stack once.
**
** If there is a destructor routine associated with the token which
** is popped from the stack, then call it.
*/
static void yy_pop_parser_stack(yyParser *pParser){
  yyStackEntry *yytos;
  assert( pParser->yytos!=0 );
  assert( pParser->yytos > pParser->yystack );
  yytos = pParser->yytos--;
#ifndef NDEBUG
  if( yyTraceFILE ){
    fprintf(yyTraceFILE,"%sPopping %s\n",
      yyTracePrompt,
      yyTokenName[yytos->major]);
  }
#endif
  yy_destructor(pParser, yytos->major, &yytos->minor);
}

/*
** Clear all secondary memory allocations from the parser
*/
static void MateParserFinalize(void *p){
  yyParser *pParser = (yyParser*)p;
  while( pParser->yytos>pParser->yystack ) yy_pop_parser_stack(pParser);
#if YYSTACKDEPTH<=0
  if( pParser->yystack!=&pParser->yystk0 ) free(pParser->yystack);
#endif
}

#ifndef MateParser_ENGINEALWAYSONSTACK
/*
** Deallocate and destroy a parser.  Destructors are called for
** all stack elements before shutting the parser down.
**
** If the YYPARSEFREENEVERNULL macro exists (for example because it
** is defined in a %include section of the input grammar) then it is
** assumed that the input pointer is never NULL.
*/
void MateParserFree(
  void *p,                    /* The parser to be deleted */
  void (*freeProc)(void*)     /* Function used to reclaim memory */
){
#ifndef YYPARSEFREENEVERNULL
  if( p==0 ) return;
#endif
  MateParserFinalize(p);
  (*freeProc)(p);
}
#endif /* MateParser_ENGINEALWAYSONSTACK */

/*
** Return the peak depth of the stack for a parser.
*/
#ifdef YYTRACKMAXSTACKDEPTH
int MateParserStackPeak(void *p){
  yyParser *pParser = (yyParser*)p;
  return pParser->yyhwm;
}
#endif

/*
** Find the appropriate action for a parser given the terminal
** look-ahead token iLookAhead.
*/
static unsigned int yy_find_shift_action(
  yyParser *pParser,        /* The parser */
  YYCODETYPE iLookAhead     /* The look-ahead token */
){
  int i;
  int stateno = pParser->yytos->stateno;

  if( stateno>=YY_MIN_REDUCE ) return stateno;
  assert( stateno <= YY_SHIFT_COUNT );
  do{
    i = yy_shift_ofst[stateno];
    assert( iLookAhead!=YYNOCODE );
    i += iLookAhead;
    if( i<0 || i>=YY_ACTTAB_COUNT || yy_lookahead[i]!=iLookAhead ){
#ifdef YYFALLBACK
      YYCODETYPE iFallback;            /* Fallback token */
      if( iLookAhead<sizeof(yyFallback)/sizeof(yyFallback[0])
             && (iFallback = yyFallback[iLookAhead])!=0 ){
#ifndef NDEBUG
        if( yyTraceFILE ){
          fprintf(yyTraceFILE, "%sFALLBACK %s => %s\n",
             yyTracePrompt, yyTokenName[iLookAhead], yyTokenName[iFallback]);
        }
#endif
        assert( yyFallback[iFallback]==0 ); /* Fallback loop must terminate */
        iLookAhead = iFallback;
        continue;
      }
#endif
#ifdef YYWILDCARD
      {
        int j = i - iLookAhead + YYWILDCARD;
        if(
#if YY_SHIFT_MIN+YYWILDCARD<0
          j>=0 &&
#endif
#if YY_SHIFT_MAX+YYWILDCARD>=YY_ACTTAB_COUNT
          j<YY_ACTTAB_COUNT &&
#endif
          yy_lookahead[j]==YYWILDCARD && iLookAhead>0
        ){
#ifndef NDEBUG
          if( yyTraceFILE ){
            fprintf(yyTraceFILE, "%sWILDCARD %s => %s\n",
               yyTracePrompt, yyTokenName[iLookAhead],
               yyTokenName[YYWILDCARD]);
          }
#endif /* NDEBUG */
          return yy_action[j];
        }
      }
#endif /* YYWILDCARD */
      return yy_default[stateno];
    }else{
      return yy_action[i];
    }
  }while(1);
}

/*
** Find the appropriate action for a parser given the non-terminal
** look-ahead token iLookAhead.
*/
static int yy_find_reduce_action(
  int stateno,              /* Current state number */
  YYCODETYPE iLookAhead     /* The look-ahead token */
){
  int i;
#ifdef YYERRORSYMBOL
  if( stateno>YY_REDUCE_COUNT ){
    return yy_default[stateno];
  }
#else
  assert( stateno<=YY_REDUCE_COUNT );
#endif
  i = yy_reduce_ofst[stateno];
  assert( i!=YY_REDUCE_USE_DFLT );
  assert( iLookAhead!=YYNOCODE );
  i += iLookAhead;
#ifdef YYERRORSYMBOL
  if( i<0 || i>=YY_ACTTAB_COUNT || yy_lookahead[i]!=iLookAhead ){
    return yy_default[stateno];
  }
#else
  assert( i>=0 && i<YY_ACTTAB_COUNT );
  assert( yy_lookahead[i]==iLookAhead );
#endif
  return yy_action[i];
}

/*
** The following routine is called if the stack overflows.
*/
static void yyStackOverflow(yyParser *yypParser){
   MateParserARG_FETCH;
#ifndef NDEBUG
   if( yyTraceFILE ){
     fprintf(yyTraceFILE,"%sStack Overflow!\n",yyTracePrompt);
   }
#endif
   while( yypParser->yytos>yypParser->yystack ) yy_pop_parser_stack(yypParser);
   /* Here code is inserted which will execute if the parser
   ** stack every overflows */
/******** Begin %stack_overflow code ******************************************/
/******** End %stack_overflow code ********************************************/
   MateParserARG_STORE; /* Suppress warning about unused %extra_argument var */
}

/*
** Print tracing information for a SHIFT action
*/
#ifndef NDEBUG
static void yyTraceShift(yyParser *yypParser, int yyNewState){
  if( yyTraceFILE ){
    if( yyNewState<YYNSTATE ){
      fprintf(yyTraceFILE,"%sShift '%s', go to state %d\n",
         yyTracePrompt,yyTokenName[yypParser->yytos->major],
         yyNewState);
    }else{
      fprintf(yyTraceFILE,"%sShift '%s'\n",
         yyTracePrompt,yyTokenName[yypParser->yytos->major]);
    }
  }
}
#else
# define yyTraceShift(X,Y)
#endif

/*
** Perform a shift action.
*/
static void yy_shift(
  yyParser *yypParser,          /* The parser to be shifted */
  int yyNewState,               /* The new state to shift in */
  int yyMajor,                  /* The major token to shift in */
  MateParserTOKENTYPE yyMinor        /* The minor token to shift in */
){
  yyStackEntry *yytos;
  yypParser->yytos++;
#ifdef YYTRACKMAXSTACKDEPTH
  if( (int)(yypParser->yytos - yypParser->yystack)>yypParser->yyhwm ){
    yypParser->yyhwm++;
    assert( yypParser->yyhwm == (int)(yypParser->yytos - yypParser->yystack) );
  }
#endif
#if YYSTACKDEPTH>0
  if( yypParser->yytos>=&yypParser->yystack[YYSTACKDEPTH] ){
    yypParser->yytos--;
    yyStackOverflow(yypParser);
    return;
  }
#else
  if( yypParser->yytos>=&yypParser->yystack[yypParser->yystksz] ){
    if( yyGrowStack(yypParser) ){
      yypParser->yytos--;
      yyStackOverflow(yypParser);
      return;
    }
  }
#endif
  if( yyNewState > YY_MAX_SHIFT ){
    yyNewState += YY_MIN_REDUCE - YY_MIN_SHIFTREDUCE;
  }
  yytos = yypParser->yytos;
  yytos->stateno = (YYACTIONTYPE)yyNewState;
  yytos->major = (YYCODETYPE)yyMajor;
  yytos->minor.yy0 = yyMinor;
  yyTraceShift(yypParser, yyNewState);
}

/* The following table contains information about every rule that
** is used during the reduce.
*/
static const struct {
  YYCODETYPE lhs;         /* Symbol on the left-hand side of the rule */
  unsigned char nrhs;     /* Number of right-hand side symbols in the rule */
} yyRuleInfo[] = {
  { 118, 3 },
  { 118, 3 },
  { 119, 3 },
  { 120, 4 },
  { 121, 4 },
  { 122, 4 },
  { 126, 3 },
  { 127, 3 },
  { 128, 3 },
  { 129, 3 },
  { 130, 3 },
  { 131, 3 },
  { 132, 3 },
  { 133, 3 },
  { 134, 3 },
  { 135, 3 },
  { 136, 3 },
  { 137, 3 },
  { 63, 4 },
  { 64, 3 },
  { 65, 2 },
  { 65, 1 },
  { 66, 3 },
  { 67, 3 },
  { 67, 0 },
  { 68, 0 },
  { 68, 2 },
  { 69, 0 },
  { 69, 1 },
  { 69, 1 },
  { 69, 1 },
  { 70, 1 },
  { 70, 1 },
  { 70, 0 },
  { 113, 16 },
  { 101, 0 },
  { 101, 3 },
  { 76, 0 },
  { 76, 5 },
  { 77, 0 },
  { 77, 1 },
  { 77, 1 },
  { 82, 2 },
  { 82, 1 },
  { 81, 5 },
  { 78, 3 },
  { 78, 0 },
  { 79, 3 },
  { 79, 0 },
  { 80, 3 },
  { 80, 0 },
  { 102, 3 },
  { 102, 1 },
  { 103, 1 },
  { 114, 19 },
  { 87, 3 },
  { 87, 0 },
  { 84, 3 },
  { 84, 0 },
  { 85, 3 },
  { 85, 0 },
  { 88, 3 },
  { 88, 0 },
  { 89, 3 },
  { 89, 0 },
  { 90, 3 },
  { 90, 0 },
  { 91, 3 },
  { 91, 0 },
  { 92, 3 },
  { 92, 0 },
  { 74, 1 },
  { 74, 1 },
  { 74, 1 },
  { 74, 1 },
  { 75, 1 },
  { 75, 1 },
  { 73, 1 },
  { 72, 1 },
  { 72, 1 },
  { 115, 11 },
  { 95, 3 },
  { 95, 0 },
  { 94, 3 },
  { 94, 0 },
  { 100, 1 },
  { 100, 1 },
  { 100, 1 },
  { 96, 2 },
  { 96, 1 },
  { 97, 4 },
  { 71, 1 },
  { 86, 3 },
  { 86, 0 },
  { 98, 3 },
  { 98, 0 },
  { 104, 3 },
  { 104, 1 },
  { 99, 1 },
  { 105, 3 },
  { 105, 2 },
  { 106, 3 },
  { 106, 1 },
  { 107, 3 },
  { 107, 1 },
  { 107, 4 },
  { 109, 3 },
  { 109, 1 },
  { 108, 1 },
  { 108, 1 },
  { 108, 1 },
  { 108, 1 },
  { 108, 1 },
  { 108, 1 },
  { 110, 1 },
  { 111, 2 },
  { 111, 0 },
  { 112, 1 },
  { 112, 1 },
  { 112, 1 },
  { 112, 1 },
  { 112, 1 },
  { 112, 1 },
  { 112, 2 },
  { 117, 9 },
  { 118, 0 },
  { 119, 0 },
  { 120, 0 },
  { 121, 0 },
  { 122, 0 },
  { 116, 7 },
  { 123, 7 },
  { 123, 0 },
  { 126, 0 },
  { 127, 0 },
  { 128, 0 },
  { 124, 10 },
  { 124, 0 },
  { 129, 0 },
  { 130, 0 },
  { 131, 0 },
  { 132, 0 },
  { 133, 0 },
  { 134, 0 },
  { 125, 8 },
  { 125, 0 },
  { 135, 0 },
  { 136, 0 },
  { 137, 0 },
};

static void yy_accept(yyParser*);  /* Forward Declaration */

/*
** Perform a reduce action and the shift that must immediately
** follow the reduce.
*/
static void yy_reduce(
  yyParser *yypParser,         /* The parser */
  unsigned int yyruleno        /* Number of the rule by which to reduce */
){
  int yygoto;                     /* The next state */
  int yyact;                      /* The next action */
  yyStackEntry *yymsp;            /* The top of the parser's stack */
  int yysize;                     /* Amount to pop the stack */
  MateParserARG_FETCH;
  yymsp = yypParser->yytos;
#ifndef NDEBUG
  if( yyTraceFILE && yyruleno<(int)(sizeof(yyRuleName)/sizeof(yyRuleName[0])) ){
    yysize = yyRuleInfo[yyruleno].nrhs;
    fprintf(yyTraceFILE, "%sReduce [%s], go to state %d.\n", yyTracePrompt,
      yyRuleName[yyruleno], yymsp[-yysize].stateno);
  }
#endif /* NDEBUG */

  /* Check that the stack is large enough to grow by a single entry
  ** if the RHS of the rule is empty.  This ensures that there is room
  ** enough on the stack to push the LHS value */
  if( yyRuleInfo[yyruleno].nrhs==0 ){
#ifdef YYTRACKMAXSTACKDEPTH
    if( (int)(yypParser->yytos - yypParser->yystack)>yypParser->yyhwm ){
      yypParser->yyhwm++;
      assert( yypParser->yyhwm == (int)(yypParser->yytos - yypParser->yystack));
    }
#endif
#if YYSTACKDEPTH>0
    if( yypParser->yytos>=&yypParser->yystack[YYSTACKDEPTH-1] ){
      yyStackOverflow(yypParser);
      return;
    }
#else
    if( yypParser->yytos>=&yypParser->yystack[yypParser->yystksz-1] ){
      if( yyGrowStack(yypParser) ){
        yyStackOverflow(yypParser);
        return;
      }
      yymsp = yypParser->yytos;
    }
#endif
  }

  switch( yyruleno ){
  /* Beginning here are the reduction cases.  A typical example
  ** follows:
  **   case 0:
  **  #line <lineno> <grammarfile>
  **     { ... }           // User supplied code
  **  #line <lineno> <thisfile>
  **     break;
  */
/********** Begin reduce actions **********************************************/
        YYMINORTYPE yylhsminor;
      case 0: /* dbgfile_default ::= FILENAME_KW QUOTED SEMICOLON */
{  yy_destructor(yypParser,6,&yymsp[-2].minor);
#line 277 "mate_grammar.lemon"
{ mc->dbg_facility = ws_fopen(yymsp[-1].minor.yy0,"w"); if (mc->dbg_facility == NULL) report_open_failure(yymsp[-1].minor.yy0,errno,TRUE); }
#line 1459 "mate_grammar.c"
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 1: /* dbgfile_default ::= FILENAME_KW NAME SEMICOLON */
{  yy_destructor(yypParser,6,&yymsp[-2].minor);
#line 278 "mate_grammar.lemon"
{ mc->dbg_facility = ws_fopen(yymsp[-1].minor.yy0,"w"); if (mc->dbg_facility == NULL) report_open_failure(yymsp[-1].minor.yy0,errno,TRUE);  }
#line 1467 "mate_grammar.c"
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 2: /* dbglevel_default ::= LEVEL_KW INTEGER SEMICOLON */
{  yy_destructor(yypParser,9,&yymsp[-2].minor);
#line 281 "mate_grammar.lemon"
{ mc->dbg_lvl = (int) strtol(yymsp[-1].minor.yy0,NULL,10); }
#line 1475 "mate_grammar.c"
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 3: /* pdu_dbglevel_default ::= PDU_KW LEVEL_KW INTEGER SEMICOLON */
{  yy_destructor(yypParser,11,&yymsp[-3].minor);
#line 284 "mate_grammar.lemon"
{ mc->dbg_pdu_lvl = (int) strtol(yymsp[-1].minor.yy0,NULL,10); }
#line 1483 "mate_grammar.c"
  yy_destructor(yypParser,9,&yymsp[-2].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 4: /* gop_dbglevel_default ::= GOP_KW LEVEL_KW INTEGER SEMICOLON */
{  yy_destructor(yypParser,12,&yymsp[-3].minor);
#line 287 "mate_grammar.lemon"
{ mc->dbg_gop_lvl = (int) strtol(yymsp[-1].minor.yy0,NULL,10); }
#line 1492 "mate_grammar.c"
  yy_destructor(yypParser,9,&yymsp[-2].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 5: /* gog_dbglevel_default ::= GOG_KW LEVEL_KW INTEGER SEMICOLON */
{  yy_destructor(yypParser,13,&yymsp[-3].minor);
#line 290 "mate_grammar.lemon"
{ mc->dbg_gog_lvl = (int) strtol(yymsp[-1].minor.yy0,NULL,10); }
#line 1501 "mate_grammar.c"
  yy_destructor(yypParser,9,&yymsp[-2].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 6: /* pdu_last_extracted_default ::= LAST_EXTRACTED_KW true_false SEMICOLON */
{  yy_destructor(yypParser,15,&yymsp[-2].minor);
#line 302 "mate_grammar.lemon"
{ mc->defaults.pdu.last_extracted = yymsp[-1].minor.yy216; }
#line 1510 "mate_grammar.c"
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 7: /* pdu_drop_unassigned_default ::= DROP_UNASSIGNED_KW true_false SEMICOLON */
{  yy_destructor(yypParser,16,&yymsp[-2].minor);
#line 305 "mate_grammar.lemon"
{ mc->defaults.pdu.drop_unassigned = yymsp[-1].minor.yy216; }
#line 1518 "mate_grammar.c"
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 8: /* pdu_discard_default ::= DISCARD_PDU_DATA_KW true_false SEMICOLON */
{  yy_destructor(yypParser,17,&yymsp[-2].minor);
#line 308 "mate_grammar.lemon"
{ mc->defaults.pdu.discard = yymsp[-1].minor.yy216; }
#line 1526 "mate_grammar.c"
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 9: /* gop_expiration_default ::= EXPIRATION_KW time_value SEMICOLON */
      case 15: /* gog_expiration_default ::= EXPIRATION_KW time_value SEMICOLON */ yytestcase(yyruleno==15);
{  yy_destructor(yypParser,18,&yymsp[-2].minor);
#line 314 "mate_grammar.lemon"
{ mc->defaults.gop.expiration = yymsp[-1].minor.yy83; }
#line 1535 "mate_grammar.c"
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 10: /* gop_idle_timeout_default ::= IDLE_TIMEOUT_KW time_value SEMICOLON */
{  yy_destructor(yypParser,19,&yymsp[-2].minor);
#line 317 "mate_grammar.lemon"
{ mc->defaults.gop.idle_timeout = yymsp[-1].minor.yy83; }
#line 1543 "mate_grammar.c"
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 11: /* gop_lifetime_default ::= LIFETIME_KW time_value SEMICOLON */
{  yy_destructor(yypParser,20,&yymsp[-2].minor);
#line 320 "mate_grammar.lemon"
{ mc->defaults.gop.lifetime = yymsp[-1].minor.yy83; }
#line 1551 "mate_grammar.c"
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 12: /* gop_drop_unassigned_default ::= DROP_UNASSIGNED_KW true_false SEMICOLON */
{  yy_destructor(yypParser,16,&yymsp[-2].minor);
#line 323 "mate_grammar.lemon"
{ mc->defaults.gop.drop_unassigned = yymsp[-1].minor.yy216; }
#line 1559 "mate_grammar.c"
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 13: /* gop_tree_mode_default ::= SHOW_TREE_KW gop_tree_mode SEMICOLON */
{  yy_destructor(yypParser,21,&yymsp[-2].minor);
#line 326 "mate_grammar.lemon"
{ mc->defaults.gop.pdu_tree_mode = (gop_pdu_tree_t)yymsp[-1].minor.yy93; }
#line 1567 "mate_grammar.c"
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 14: /* gop_show_times_default ::= SHOW_TIMES_KW true_false SEMICOLON */
{  yy_destructor(yypParser,22,&yymsp[-2].minor);
#line 329 "mate_grammar.lemon"
{ mc->defaults.gop.show_times = yymsp[-1].minor.yy216; }
#line 1575 "mate_grammar.c"
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 16: /* gog_goptree_default ::= GOP_TREE_KW gop_tree_type SEMICOLON */
{  yy_destructor(yypParser,23,&yymsp[-2].minor);
#line 338 "mate_grammar.lemon"
{ mc->defaults.gog.gop_tree_mode = yymsp[-1].minor.yy93; }
#line 1583 "mate_grammar.c"
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 17: /* gog_show_times_default ::= SHOW_TIMES_KW true_false SEMICOLON */
{  yy_destructor(yypParser,22,&yymsp[-2].minor);
#line 341 "mate_grammar.lemon"
{ mc->defaults.gog.show_times = yymsp[-1].minor.yy216; }
#line 1591 "mate_grammar.c"
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 18: /* transform_decl ::= TRANSFORM_KW NAME transform_body SEMICOLON */
{  yy_destructor(yypParser,24,&yymsp[-3].minor);
#line 348 "mate_grammar.lemon"
{
	AVPL_Transf* c;

	if ( g_hash_table_lookup(mc->transfs,yymsp[-2].minor.yy0) ) {
		configuration_error(mc,"yymsp[-3].minor.yy71 transformation called '%s' exists already",yymsp[-2].minor.yy0);
	}

	for ( c = yymsp[-1].minor.yy71; c; c = c->next )
		c->name = g_strdup(yymsp[-2].minor.yy0);

	g_hash_table_insert(mc->transfs,yymsp[-1].minor.yy71->name,yymsp[-1].minor.yy71);

	yymsp[-3].minor.yy71 = NULL;
}
#line 1612 "mate_grammar.c"
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 19: /* transform_body ::= OPEN_BRACE transform_statements CLOSE_BRACE */
{  yy_destructor(yypParser,4,&yymsp[-2].minor);
#line 363 "mate_grammar.lemon"
{ yymsp[-2].minor.yy71 = yymsp[-1].minor.yy71; }
#line 1620 "mate_grammar.c"
  yy_destructor(yypParser,5,&yymsp[0].minor);
}
        break;
      case 20: /* transform_statements ::= transform_statements transform_statement */
#line 365 "mate_grammar.lemon"
{
	AVPL_Transf* c;

	for ( c = yymsp[-1].minor.yy71; c->next; c = c->next ) ;
	c->next = yymsp[0].minor.yy71;
	yylhsminor.yy71 = yymsp[-1].minor.yy71;
}
#line 1633 "mate_grammar.c"
  yymsp[-1].minor.yy71 = yylhsminor.yy71;
        break;
      case 21: /* transform_statements ::= transform_statement */
#line 373 "mate_grammar.lemon"
{ yylhsminor.yy71 = yymsp[0].minor.yy71; }
#line 1639 "mate_grammar.c"
  yymsp[0].minor.yy71 = yylhsminor.yy71;
        break;
      case 22: /* transform_statement ::= transform_match transform_action SEMICOLON */
#line 375 "mate_grammar.lemon"
{
	yylhsminor.yy71 = new_transform_elem(yymsp[-2].minor.yy16->avpl,yymsp[-1].minor.yy77->avpl,yymsp[-2].minor.yy16->match_mode,yymsp[-1].minor.yy77->replace_mode);
}
#line 1647 "mate_grammar.c"
  yy_destructor(yypParser,2,&yymsp[0].minor);
  yymsp[-2].minor.yy71 = yylhsminor.yy71;
        break;
      case 23: /* transform_match ::= MATCH_KW match_mode avpl */
{  yy_destructor(yypParser,25,&yymsp[-2].minor);
#line 379 "mate_grammar.lemon"
{
	yymsp[-2].minor.yy16 = (transf_match_t *)g_malloc(sizeof(transf_match_t));
	yymsp[-2].minor.yy16->match_mode = yymsp[-1].minor.yy172;
	yymsp[-2].minor.yy16->avpl = yymsp[0].minor.yy170;
}
#line 1659 "mate_grammar.c"
}
        break;
      case 24: /* transform_match ::= */
#line 385 "mate_grammar.lemon"
{
	yymsp[1].minor.yy16 = (transf_match_t *)g_malloc(sizeof(transf_match_t));
	yymsp[1].minor.yy16->match_mode = AVPL_STRICT;
	yymsp[1].minor.yy16->avpl = new_avpl("");

}
#line 1670 "mate_grammar.c"
        break;
      case 25: /* transform_action ::= */
#line 392 "mate_grammar.lemon"
{
	yymsp[1].minor.yy77 = (transf_action_t *)g_malloc(sizeof(transf_action_t));
	yymsp[1].minor.yy77->replace_mode = AVPL_INSERT;
	yymsp[1].minor.yy77->avpl = new_avpl("");
}
#line 1679 "mate_grammar.c"
        break;
      case 26: /* transform_action ::= action_mode avpl */
#line 397 "mate_grammar.lemon"
{
	yylhsminor.yy77 = (transf_action_t *)g_malloc(sizeof(transf_action_t));
	yylhsminor.yy77->replace_mode = yymsp[-1].minor.yy215;
	yylhsminor.yy77->avpl = yymsp[0].minor.yy170;
}
#line 1688 "mate_grammar.c"
  yymsp[-1].minor.yy77 = yylhsminor.yy77;
        break;
      case 27: /* match_mode ::= */
#line 403 "mate_grammar.lemon"
{ yymsp[1].minor.yy172 = AVPL_STRICT; }
#line 1694 "mate_grammar.c"
        break;
      case 28: /* match_mode ::= STRICT_KW */
{  yy_destructor(yypParser,26,&yymsp[0].minor);
#line 404 "mate_grammar.lemon"
{ yymsp[0].minor.yy172 = AVPL_STRICT; }
#line 1700 "mate_grammar.c"
}
        break;
      case 29: /* match_mode ::= EVERY_KW */
{  yy_destructor(yypParser,27,&yymsp[0].minor);
#line 405 "mate_grammar.lemon"
{ yymsp[0].minor.yy172 = AVPL_EVERY; }
#line 1707 "mate_grammar.c"
}
        break;
      case 30: /* match_mode ::= LOOSE_KW */
{  yy_destructor(yypParser,28,&yymsp[0].minor);
#line 406 "mate_grammar.lemon"
{ yymsp[0].minor.yy172 = AVPL_LOOSE; }
#line 1714 "mate_grammar.c"
}
        break;
      case 31: /* action_mode ::= REPLACE_KW */
{  yy_destructor(yypParser,29,&yymsp[0].minor);
#line 408 "mate_grammar.lemon"
{ yymsp[0].minor.yy215 = AVPL_REPLACE; }
#line 1721 "mate_grammar.c"
}
        break;
      case 32: /* action_mode ::= INSERT_KW */
{  yy_destructor(yypParser,30,&yymsp[0].minor);
#line 409 "mate_grammar.lemon"
{ yymsp[0].minor.yy215 = AVPL_INSERT; }
#line 1728 "mate_grammar.c"
}
        break;
      case 33: /* action_mode ::= */
#line 410 "mate_grammar.lemon"
{ yymsp[1].minor.yy215 = AVPL_INSERT; }
#line 1734 "mate_grammar.c"
        break;
      case 34: /* pdu_decl ::= PDU_KW NAME PROTO_KW field TRANSPORT_KW proto_stack OPEN_BRACE payload_statement extraction_statements transform_list_statement criteria_statement pdu_drop_unassigned_statement discard_pdu_data_statement last_extracted_statement CLOSE_BRACE SEMICOLON */
{  yy_destructor(yypParser,11,&yymsp[-15].minor);
#line 426 "mate_grammar.lemon"
{

	mate_cfg_pdu* cfg  = new_pducfg(mc, yymsp[-14].minor.yy0);
	extraction_t *extraction, *next_extraction;
	GPtrArray* transport_stack = g_ptr_array_new();
	int i;

	if (! cfg ) configuration_error(mc,"could not create Pdu %s.",yymsp[-14].minor.yy0);

	cfg->hfid_proto = yymsp[-12].minor.yy202->id;

	cfg->last_extracted = yymsp[-2].minor.yy216;
	cfg->discard = yymsp[-3].minor.yy216;
	cfg->drop_unassigned = yymsp[-4].minor.yy216;

	/*
	 * Add this protocol to our table of wanted hfids.
	 */
	mc->wanted_hfids = g_array_append_val(mc->wanted_hfids, yymsp[-12].minor.yy202->id);

	/* flip the transport_stack */
	for (i = yymsp[-10].minor.yy231->len - 1; yymsp[-10].minor.yy231->len; i--) {
		g_ptr_array_add(transport_stack,g_ptr_array_remove_index(yymsp[-10].minor.yy231,i));
	}

	g_ptr_array_free(yymsp[-10].minor.yy231,FALSE);

	cfg->transport_ranges = transport_stack;
	cfg->payload_ranges = yymsp[-8].minor.yy231;

	if (yymsp[-5].minor.yy41) {
		cfg->criterium = yymsp[-5].minor.yy41->criterium_avpl;
		cfg->criterium_match_mode = yymsp[-5].minor.yy41->criterium_match_mode;
		cfg->criterium_accept_mode = yymsp[-5].minor.yy41->criterium_accept_mode;
	}

	cfg->transforms = yymsp[-6].minor.yy231;

	for (extraction = yymsp[-7].minor.yy259; extraction; extraction = next_extraction) {
		next_extraction = extraction->next;

		if ( ! add_hfid(mc, extraction->hfi, extraction->as, cfg->hfids_attr) ) {
			configuration_error(mc,"MATE: failed to create extraction rule '%s'",extraction->as);
		}

		g_free(extraction);
	}
}
#line 1787 "mate_grammar.c"
  yy_destructor(yypParser,31,&yymsp[-13].minor);
  yy_destructor(yypParser,32,&yymsp[-11].minor);
  yy_destructor(yypParser,4,&yymsp[-9].minor);
  yy_destructor(yypParser,5,&yymsp[-1].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 35: /* payload_statement ::= */
#line 475 "mate_grammar.lemon"
{ yymsp[1].minor.yy231 = NULL; }
#line 1798 "mate_grammar.c"
        break;
      case 36: /* payload_statement ::= PAYLOAD_KW proto_stack SEMICOLON */
{  yy_destructor(yypParser,33,&yymsp[-2].minor);
#line 476 "mate_grammar.lemon"
{ yymsp[-2].minor.yy231 = yymsp[-1].minor.yy231; }
#line 1804 "mate_grammar.c"
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 37: /* criteria_statement ::= */
#line 478 "mate_grammar.lemon"
{ yymsp[1].minor.yy41 = NULL; }
#line 1811 "mate_grammar.c"
        break;
      case 38: /* criteria_statement ::= CRITERIA_KW accept_mode match_mode avpl SEMICOLON */
{  yy_destructor(yypParser,34,&yymsp[-4].minor);
#line 479 "mate_grammar.lemon"
{
	yymsp[-4].minor.yy41 = (pdu_criteria_t *)g_malloc(sizeof(pdu_criteria_t));
	yymsp[-4].minor.yy41->criterium_avpl = yymsp[-1].minor.yy170;
	yymsp[-4].minor.yy41->criterium_match_mode = yymsp[-2].minor.yy172;
	yymsp[-4].minor.yy41->criterium_accept_mode = yymsp[-3].minor.yy188;
}
#line 1822 "mate_grammar.c"
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 39: /* accept_mode ::= */
#line 486 "mate_grammar.lemon"
{ yymsp[1].minor.yy188 = ACCEPT_MODE; }
#line 1829 "mate_grammar.c"
        break;
      case 40: /* accept_mode ::= ACCEPT_KW */
{  yy_destructor(yypParser,35,&yymsp[0].minor);
#line 487 "mate_grammar.lemon"
{ yymsp[0].minor.yy188 = ACCEPT_MODE; }
#line 1835 "mate_grammar.c"
}
        break;
      case 41: /* accept_mode ::= REJECT_KW */
{  yy_destructor(yypParser,36,&yymsp[0].minor);
#line 488 "mate_grammar.lemon"
{ yymsp[0].minor.yy188 = REJECT_MODE; }
#line 1842 "mate_grammar.c"
}
        break;
      case 42: /* extraction_statements ::= extraction_statements extraction_statement */
#line 490 "mate_grammar.lemon"
{ yylhsminor.yy259 = yymsp[-1].minor.yy259; yylhsminor.yy259->last = yylhsminor.yy259->last->next = yymsp[0].minor.yy259; }
#line 1848 "mate_grammar.c"
  yymsp[-1].minor.yy259 = yylhsminor.yy259;
        break;
      case 43: /* extraction_statements ::= extraction_statement */
#line 491 "mate_grammar.lemon"
{ yylhsminor.yy259 = yymsp[0].minor.yy259; yylhsminor.yy259->last = yylhsminor.yy259; }
#line 1854 "mate_grammar.c"
  yymsp[0].minor.yy259 = yylhsminor.yy259;
        break;
      case 44: /* extraction_statement ::= EXTRACT_KW NAME FROM_KW field SEMICOLON */
{  yy_destructor(yypParser,37,&yymsp[-4].minor);
#line 493 "mate_grammar.lemon"
{
	yymsp[-4].minor.yy259 = (extraction_t *)g_malloc(sizeof(extraction_t));
	yymsp[-4].minor.yy259->as = yymsp[-3].minor.yy0;
	yymsp[-4].minor.yy259->hfi = yymsp[-1].minor.yy202;
	yymsp[-4].minor.yy259->next = yymsp[-4].minor.yy259->last = NULL;
}
#line 1866 "mate_grammar.c"
  yy_destructor(yypParser,38,&yymsp[-2].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 45: /* pdu_drop_unassigned_statement ::= DROP_UNASSIGNED_KW true_false SEMICOLON */
      case 55: /* gop_drop_unassigned_statement ::= DROP_UNASSIGNED_KW true_false SEMICOLON */ yytestcase(yyruleno==55);
{  yy_destructor(yypParser,16,&yymsp[-2].minor);
#line 501 "mate_grammar.lemon"
{ yymsp[-2].minor.yy216 = yymsp[-1].minor.yy216; }
#line 1876 "mate_grammar.c"
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 46: /* pdu_drop_unassigned_statement ::= */
#line 502 "mate_grammar.lemon"
{ yymsp[1].minor.yy216 =  mc->defaults.pdu.drop_unassigned; }
#line 1883 "mate_grammar.c"
        break;
      case 47: /* discard_pdu_data_statement ::= DISCARD_PDU_DATA_KW true_false SEMICOLON */
{  yy_destructor(yypParser,17,&yymsp[-2].minor);
#line 504 "mate_grammar.lemon"
{ yymsp[-2].minor.yy216 = yymsp[-1].minor.yy216; }
#line 1889 "mate_grammar.c"
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 48: /* discard_pdu_data_statement ::= */
#line 505 "mate_grammar.lemon"
{ yymsp[1].minor.yy216 =  mc->defaults.pdu.discard; }
#line 1896 "mate_grammar.c"
        break;
      case 49: /* last_extracted_statement ::= LAST_PDU_KW true_false SEMICOLON */
{  yy_destructor(yypParser,39,&yymsp[-2].minor);
#line 507 "mate_grammar.lemon"
{ yymsp[-2].minor.yy216 = yymsp[-1].minor.yy216; }
#line 1902 "mate_grammar.c"
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 50: /* last_extracted_statement ::= */
#line 508 "mate_grammar.lemon"
{ yymsp[1].minor.yy216 = mc->defaults.pdu.last_extracted; }
#line 1909 "mate_grammar.c"
        break;
      case 51: /* proto_stack ::= proto_stack SLASH field */
#line 510 "mate_grammar.lemon"
{
	int* hfidp = (int *)g_malloc(sizeof(int));

	*hfidp = yymsp[0].minor.yy202->id;
	g_ptr_array_add(yymsp[-2].minor.yy231,hfidp);
	yylhsminor.yy231 = yymsp[-2].minor.yy231;
}
#line 1920 "mate_grammar.c"
  yy_destructor(yypParser,40,&yymsp[-1].minor);
  yymsp[-2].minor.yy231 = yylhsminor.yy231;
        break;
      case 52: /* proto_stack ::= field */
#line 518 "mate_grammar.lemon"
{
	int* hfidp = (int *)g_malloc(sizeof(int));
	*hfidp = yymsp[0].minor.yy202->id;

	yylhsminor.yy231 = g_ptr_array_new();
	g_ptr_array_add(yylhsminor.yy231,hfidp);
}
#line 1933 "mate_grammar.c"
  yymsp[0].minor.yy231 = yylhsminor.yy231;
        break;
      case 53: /* field ::= NAME */
#line 526 "mate_grammar.lemon"
{
	yylhsminor.yy202 = proto_registrar_get_byname(yymsp[0].minor.yy0);
}
#line 1941 "mate_grammar.c"
  yymsp[0].minor.yy202 = yylhsminor.yy202;
        break;
      case 54: /* gop_decl ::= GOP_KW NAME ON_KW pdu_name MATCH_KW avpl OPEN_BRACE gop_start_statement gop_stop_statement extra_statement transform_list_statement gop_expiration_statement idle_timeout_statement lifetime_statement gop_drop_unassigned_statement show_goptree_statement show_times_statement CLOSE_BRACE SEMICOLON */
{  yy_destructor(yypParser,12,&yymsp[-18].minor);
#line 544 "mate_grammar.lemon"
{
		mate_cfg_gop* cfg;

	if (g_hash_table_lookup(mc->gopcfgs,yymsp[-17].minor.yy0)) configuration_error(mc,"yymsp[-18].minor.yy0 Gop Named '%s' exists already.",yymsp[-17].minor.yy0);
	if (g_hash_table_lookup(mc->gops_by_pduname,yymsp[-15].minor.yy88) ) configuration_error(mc,"Gop for Pdu '%s' exists already",yymsp[-15].minor.yy88);

	cfg = new_gopcfg(mc, yymsp[-17].minor.yy0);
	g_hash_table_insert(mc->gops_by_pduname,yymsp[-15].minor.yy88,cfg);
	g_hash_table_insert(mc->gopcfgs,cfg->name,cfg);

	cfg->on_pdu = yymsp[-15].minor.yy88;
	cfg->key = yymsp[-13].minor.yy170;
	cfg->drop_unassigned = yymsp[-4].minor.yy216;
	cfg->show_times = yymsp[-2].minor.yy216;
	cfg->pdu_tree_mode = (gop_pdu_tree_t)yymsp[-3].minor.yy93;
	cfg->expiration = yymsp[-7].minor.yy83;
	cfg->idle_timeout = yymsp[-6].minor.yy83;
	cfg->lifetime = yymsp[-5].minor.yy83;
	cfg->start = yymsp[-11].minor.yy170;
	cfg->stop = yymsp[-10].minor.yy170;
	cfg->transforms = yymsp[-8].minor.yy231;

	merge_avpl(cfg->extra,yymsp[-9].minor.yy170,TRUE);
	delete_avpl(yymsp[-9].minor.yy170,TRUE);
}
#line 1972 "mate_grammar.c"
  yy_destructor(yypParser,41,&yymsp[-16].minor);
  yy_destructor(yypParser,25,&yymsp[-14].minor);
  yy_destructor(yypParser,4,&yymsp[-12].minor);
  yy_destructor(yypParser,5,&yymsp[-1].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 56: /* gop_drop_unassigned_statement ::= */
#line 571 "mate_grammar.lemon"
{ yymsp[1].minor.yy216 =  mc->defaults.gop.drop_unassigned; }
#line 1983 "mate_grammar.c"
        break;
      case 57: /* gop_start_statement ::= START_KW avpl SEMICOLON */
{  yy_destructor(yypParser,42,&yymsp[-2].minor);
#line 573 "mate_grammar.lemon"
{ yymsp[-2].minor.yy170 = yymsp[-1].minor.yy170; }
#line 1989 "mate_grammar.c"
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 58: /* gop_start_statement ::= */
      case 60: /* gop_stop_statement ::= */ yytestcase(yyruleno==60);
#line 574 "mate_grammar.lemon"
{ yymsp[1].minor.yy170 = NULL; }
#line 1997 "mate_grammar.c"
        break;
      case 59: /* gop_stop_statement ::= STOP_KW avpl SEMICOLON */
{  yy_destructor(yypParser,43,&yymsp[-2].minor);
#line 576 "mate_grammar.lemon"
{ yymsp[-2].minor.yy170 = yymsp[-1].minor.yy170; }
#line 2003 "mate_grammar.c"
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 61: /* show_goptree_statement ::= SHOW_TREE_KW gop_tree_mode SEMICOLON */
{  yy_destructor(yypParser,21,&yymsp[-2].minor);
#line 579 "mate_grammar.lemon"
{ yymsp[-2].minor.yy93 = yymsp[-1].minor.yy93; }
#line 2011 "mate_grammar.c"
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 62: /* show_goptree_statement ::= */
#line 580 "mate_grammar.lemon"
{ yymsp[1].minor.yy93 = (gop_tree_mode_t)mc->defaults.gop.pdu_tree_mode; }
#line 2018 "mate_grammar.c"
        break;
      case 63: /* show_times_statement ::= SHOW_TIMES_KW true_false SEMICOLON */
{  yy_destructor(yypParser,22,&yymsp[-2].minor);
#line 582 "mate_grammar.lemon"
{ yymsp[-2].minor.yy216 = yymsp[-1].minor.yy216; }
#line 2024 "mate_grammar.c"
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 64: /* show_times_statement ::= */
#line 583 "mate_grammar.lemon"
{ yymsp[1].minor.yy216 = mc->defaults.gop.show_times; }
#line 2031 "mate_grammar.c"
        break;
      case 65: /* gop_expiration_statement ::= EXPIRATION_KW time_value SEMICOLON */
      case 83: /* gog_expiration_statement ::= EXPIRATION_KW time_value SEMICOLON */ yytestcase(yyruleno==83);
{  yy_destructor(yypParser,18,&yymsp[-2].minor);
#line 585 "mate_grammar.lemon"
{ yymsp[-2].minor.yy83 = yymsp[-1].minor.yy83; }
#line 2038 "mate_grammar.c"
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 66: /* gop_expiration_statement ::= */
      case 68: /* idle_timeout_statement ::= */ yytestcase(yyruleno==68);
      case 70: /* lifetime_statement ::= */ yytestcase(yyruleno==70);
#line 586 "mate_grammar.lemon"
{ yymsp[1].minor.yy83 = mc->defaults.gop.lifetime; }
#line 2047 "mate_grammar.c"
        break;
      case 67: /* idle_timeout_statement ::= IDLE_TIMEOUT_KW time_value SEMICOLON */
{  yy_destructor(yypParser,19,&yymsp[-2].minor);
#line 588 "mate_grammar.lemon"
{ yymsp[-2].minor.yy83 = yymsp[-1].minor.yy83; }
#line 2053 "mate_grammar.c"
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 69: /* lifetime_statement ::= LIFETIME_KW time_value SEMICOLON */
{  yy_destructor(yypParser,20,&yymsp[-2].minor);
#line 591 "mate_grammar.lemon"
{ yymsp[-2].minor.yy83 = yymsp[-1].minor.yy83; }
#line 2061 "mate_grammar.c"
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 71: /* gop_tree_mode ::= NO_TREE_KW */
{  yy_destructor(yypParser,44,&yymsp[0].minor);
#line 594 "mate_grammar.lemon"
{ yymsp[0].minor.yy93 = (gop_tree_mode_t)GOP_NO_TREE; }
#line 2069 "mate_grammar.c"
}
        break;
      case 72: /* gop_tree_mode ::= PDU_TREE_KW */
{  yy_destructor(yypParser,45,&yymsp[0].minor);
#line 595 "mate_grammar.lemon"
{ yymsp[0].minor.yy93 = (gop_tree_mode_t)GOP_PDU_TREE; }
#line 2076 "mate_grammar.c"
}
        break;
      case 73: /* gop_tree_mode ::= FRAME_TREE_KW */
{  yy_destructor(yypParser,46,&yymsp[0].minor);
#line 596 "mate_grammar.lemon"
{ yymsp[0].minor.yy93 = (gop_tree_mode_t)GOP_FRAME_TREE; }
#line 2083 "mate_grammar.c"
}
        break;
      case 74: /* gop_tree_mode ::= BASIC_TREE_KW */
{  yy_destructor(yypParser,47,&yymsp[0].minor);
#line 597 "mate_grammar.lemon"
{ yymsp[0].minor.yy93 = (gop_tree_mode_t)GOP_BASIC_PDU_TREE; }
#line 2090 "mate_grammar.c"
}
        break;
      case 75: /* true_false ::= TRUE_KW */
{  yy_destructor(yypParser,48,&yymsp[0].minor);
#line 599 "mate_grammar.lemon"
{ yymsp[0].minor.yy216 = TRUE; }
#line 2097 "mate_grammar.c"
}
        break;
      case 76: /* true_false ::= FALSE_KW */
{  yy_destructor(yypParser,49,&yymsp[0].minor);
#line 600 "mate_grammar.lemon"
{ yymsp[0].minor.yy216 = FALSE; }
#line 2104 "mate_grammar.c"
}
        break;
      case 77: /* pdu_name ::= NAME */
#line 602 "mate_grammar.lemon"
{
	mate_cfg_pdu* c;
	if (( c =  (mate_cfg_pdu *)g_hash_table_lookup(mc->pducfgs,yymsp[0].minor.yy0) )) {
		yylhsminor.yy88 = c->name;
	} else {
		configuration_error(mc,"No such Pdu: '%s'",yymsp[0].minor.yy0);
	}
}
#line 2117 "mate_grammar.c"
  yymsp[0].minor.yy88 = yylhsminor.yy88;
        break;
      case 78: /* time_value ::= FLOATING */
      case 79: /* time_value ::= INTEGER */ yytestcase(yyruleno==79);
#line 612 "mate_grammar.lemon"
{
	yylhsminor.yy83 = (float) g_ascii_strtod(yymsp[0].minor.yy0,NULL);
}
#line 2126 "mate_grammar.c"
  yymsp[0].minor.yy83 = yylhsminor.yy83;
        break;
      case 80: /* gog_decl ::= GOG_KW NAME OPEN_BRACE gog_key_statements extra_statement transform_list_statement gog_expiration_statement gog_goptree_statement show_times_statement CLOSE_BRACE SEMICOLON */
{  yy_destructor(yypParser,13,&yymsp[-10].minor);
#line 630 "mate_grammar.lemon"
{
		mate_cfg_gog* cfg = NULL;

		if ( g_hash_table_lookup(mc->gogcfgs,yymsp[-9].minor.yy0) ) {
			configuration_error(mc,"Gog '%s' exists already ",yymsp[-9].minor.yy0);
		}

		cfg = new_gogcfg(mc, yymsp[-9].minor.yy0);

		cfg->expiration = yymsp[-4].minor.yy83;
		cfg->gop_tree_mode = yymsp[-3].minor.yy93;
		cfg->transforms = yymsp[-5].minor.yy231;
		cfg->keys = yymsp[-7].minor.yy243;
		cfg->show_times = yymsp[-2].minor.yy216;

		merge_avpl(cfg->extra,yymsp[-6].minor.yy170,TRUE);
		delete_avpl(yymsp[-6].minor.yy170,TRUE);
}
#line 2150 "mate_grammar.c"
  yy_destructor(yypParser,4,&yymsp[-8].minor);
  yy_destructor(yypParser,5,&yymsp[-1].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 81: /* gog_goptree_statement ::= GOP_TREE_KW gop_tree_type SEMICOLON */
{  yy_destructor(yypParser,23,&yymsp[-2].minor);
#line 649 "mate_grammar.lemon"
{ yymsp[-2].minor.yy93 = yymsp[-1].minor.yy93; }
#line 2160 "mate_grammar.c"
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 82: /* gog_goptree_statement ::= */
#line 650 "mate_grammar.lemon"
{ yymsp[1].minor.yy93 = mc->defaults.gog.gop_tree_mode; }
#line 2167 "mate_grammar.c"
        break;
      case 84: /* gog_expiration_statement ::= */
#line 653 "mate_grammar.lemon"
{ yymsp[1].minor.yy83 = mc->defaults.gog.expiration; }
#line 2172 "mate_grammar.c"
        break;
      case 85: /* gop_tree_type ::= NULL_TREE_KW */
{  yy_destructor(yypParser,51,&yymsp[0].minor);
#line 655 "mate_grammar.lemon"
{ yymsp[0].minor.yy93 = GOP_NULL_TREE; }
#line 2178 "mate_grammar.c"
}
        break;
      case 86: /* gop_tree_type ::= FULL_TREE_KW */
{  yy_destructor(yypParser,52,&yymsp[0].minor);
#line 656 "mate_grammar.lemon"
{ yymsp[0].minor.yy93 = GOP_FULL_TREE; }
#line 2185 "mate_grammar.c"
}
        break;
      case 87: /* gop_tree_type ::= BASIC_TREE_KW */
{  yy_destructor(yypParser,47,&yymsp[0].minor);
#line 657 "mate_grammar.lemon"
{ yymsp[0].minor.yy93 = GOP_BASIC_TREE; }
#line 2192 "mate_grammar.c"
}
        break;
      case 88: /* gog_key_statements ::= gog_key_statements gog_key_statement */
#line 659 "mate_grammar.lemon"
{
	loal_append(yymsp[-1].minor.yy243,yymsp[0].minor.yy170);
	yylhsminor.yy243 = yymsp[-1].minor.yy243;
}
#line 2201 "mate_grammar.c"
  yymsp[-1].minor.yy243 = yylhsminor.yy243;
        break;
      case 89: /* gog_key_statements ::= gog_key_statement */
#line 664 "mate_grammar.lemon"
{
	yylhsminor.yy243 = new_loal("");
	loal_append(yylhsminor.yy243,yymsp[0].minor.yy170);
}
#line 2210 "mate_grammar.c"
  yymsp[0].minor.yy243 = yylhsminor.yy243;
        break;
      case 90: /* gog_key_statement ::= MEMBER_KW gop_name avpl SEMICOLON */
{  yy_destructor(yypParser,53,&yymsp[-3].minor);
#line 670 "mate_grammar.lemon"
{
	rename_avpl(yymsp[-1].minor.yy170,yymsp[-2].minor.yy88);
	yymsp[-3].minor.yy170 = yymsp[-1].minor.yy170;
}
#line 2220 "mate_grammar.c"
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 91: /* gop_name ::= NAME */
#line 675 "mate_grammar.lemon"
{
	mate_cfg_gop* c;
	if (( c = (mate_cfg_gop *)g_hash_table_lookup(mc->gopcfgs,yymsp[0].minor.yy0) )) {
		yylhsminor.yy88 = c->name;
	} else {
		configuration_error(mc,"No Gop called '%s' has been already declared",yymsp[0].minor.yy0);
	}
}
#line 2234 "mate_grammar.c"
  yymsp[0].minor.yy88 = yylhsminor.yy88;
        break;
      case 92: /* extra_statement ::= EXTRA_KW avpl SEMICOLON */
{  yy_destructor(yypParser,54,&yymsp[-2].minor);
#line 687 "mate_grammar.lemon"
{ yymsp[-2].minor.yy170 = yymsp[-1].minor.yy170; }
#line 2241 "mate_grammar.c"
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 93: /* extra_statement ::= */
#line 688 "mate_grammar.lemon"
{ yymsp[1].minor.yy170 = new_avpl(""); }
#line 2248 "mate_grammar.c"
        break;
      case 94: /* transform_list_statement ::= TRANSFORM_KW transform_list SEMICOLON */
{  yy_destructor(yypParser,24,&yymsp[-2].minor);
#line 690 "mate_grammar.lemon"
{ yymsp[-2].minor.yy231 = yymsp[-1].minor.yy231; }
#line 2254 "mate_grammar.c"
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 95: /* transform_list_statement ::= */
#line 691 "mate_grammar.lemon"
{ yymsp[1].minor.yy231 = g_ptr_array_new(); }
#line 2261 "mate_grammar.c"
        break;
      case 96: /* transform_list ::= transform_list COMMA transform */
#line 693 "mate_grammar.lemon"
{
	yylhsminor.yy231 = yymsp[-2].minor.yy231;
	g_ptr_array_add(yymsp[-2].minor.yy231,yymsp[0].minor.yy71);
}
#line 2269 "mate_grammar.c"
  yy_destructor(yypParser,55,&yymsp[-1].minor);
  yymsp[-2].minor.yy231 = yylhsminor.yy231;
        break;
      case 97: /* transform_list ::= transform */
#line 698 "mate_grammar.lemon"
{
	yylhsminor.yy231 = g_ptr_array_new();
	g_ptr_array_add(yylhsminor.yy231,yymsp[0].minor.yy71);
}
#line 2279 "mate_grammar.c"
  yymsp[0].minor.yy231 = yylhsminor.yy231;
        break;
      case 98: /* transform ::= NAME */
#line 703 "mate_grammar.lemon"
{
	AVPL_Transf* t;

	if (( t = (AVPL_Transf *)g_hash_table_lookup(mc->transfs,yymsp[0].minor.yy0) )) {
		yylhsminor.yy71 = t;
	} else {
		configuration_error(mc,"There's no such Transformation: %s",yymsp[0].minor.yy0);
	}
}
#line 2293 "mate_grammar.c"
  yymsp[0].minor.yy71 = yylhsminor.yy71;
        break;
      case 99: /* avpl ::= OPEN_PARENS avps CLOSE_PARENS */
{  yy_destructor(yypParser,56,&yymsp[-2].minor);
#line 713 "mate_grammar.lemon"
{ yymsp[-2].minor.yy170 = yymsp[-1].minor.yy170; }
#line 2300 "mate_grammar.c"
  yy_destructor(yypParser,57,&yymsp[0].minor);
}
        break;
      case 100: /* avpl ::= OPEN_PARENS CLOSE_PARENS */
{  yy_destructor(yypParser,56,&yymsp[-1].minor);
#line 714 "mate_grammar.lemon"
{ yymsp[-1].minor.yy170 = new_avpl(""); }
#line 2308 "mate_grammar.c"
  yy_destructor(yypParser,57,&yymsp[0].minor);
}
        break;
      case 101: /* avps ::= avps COMMA avp */
#line 716 "mate_grammar.lemon"
{ yylhsminor.yy170 = yymsp[-2].minor.yy170; if ( ! insert_avp(yymsp[-2].minor.yy170,yymsp[0].minor.yy110) ) delete_avp(yymsp[0].minor.yy110); }
#line 2315 "mate_grammar.c"
  yy_destructor(yypParser,55,&yymsp[-1].minor);
  yymsp[-2].minor.yy170 = yylhsminor.yy170;
        break;
      case 102: /* avps ::= avp */
#line 717 "mate_grammar.lemon"
{ yylhsminor.yy170 = new_avpl(""); if ( ! insert_avp(yylhsminor.yy170,yymsp[0].minor.yy110) ) delete_avp(yymsp[0].minor.yy110); }
#line 2322 "mate_grammar.c"
  yymsp[0].minor.yy170 = yylhsminor.yy170;
        break;
      case 103: /* avp ::= NAME AVP_OPERATOR value */
#line 719 "mate_grammar.lemon"
{ yylhsminor.yy110 = new_avp(yymsp[-2].minor.yy0,yymsp[0].minor.yy88,*yymsp[-1].minor.yy0); }
#line 2328 "mate_grammar.c"
  yymsp[-2].minor.yy110 = yylhsminor.yy110;
        break;
      case 104: /* avp ::= NAME */
#line 720 "mate_grammar.lemon"
{ yylhsminor.yy110 = new_avp(yymsp[0].minor.yy0,"",'?'); }
#line 2334 "mate_grammar.c"
  yymsp[0].minor.yy110 = yylhsminor.yy110;
        break;
      case 105: /* avp ::= NAME OPEN_BRACE avp_oneoff CLOSE_BRACE */
#line 721 "mate_grammar.lemon"
{ yylhsminor.yy110 = new_avp(yymsp[-3].minor.yy0,yymsp[-1].minor.yy88,'|'); }
#line 2340 "mate_grammar.c"
  yy_destructor(yypParser,4,&yymsp[-2].minor);
  yy_destructor(yypParser,5,&yymsp[0].minor);
  yymsp[-3].minor.yy110 = yylhsminor.yy110;
        break;
      case 106: /* avp_oneoff ::= avp_oneoff PIPE value */
#line 723 "mate_grammar.lemon"
{ yylhsminor.yy88 = g_strdup_printf("%s|%s",yymsp[-2].minor.yy88,yymsp[0].minor.yy88); }
#line 2348 "mate_grammar.c"
  yy_destructor(yypParser,59,&yymsp[-1].minor);
  yymsp[-2].minor.yy88 = yylhsminor.yy88;
        break;
      case 107: /* avp_oneoff ::= value */
#line 724 "mate_grammar.lemon"
{ yylhsminor.yy88 = g_strdup(yymsp[0].minor.yy88); }
#line 2355 "mate_grammar.c"
  yymsp[0].minor.yy88 = yylhsminor.yy88;
        break;
      case 108: /* value ::= QUOTED */
      case 109: /* value ::= NAME */ yytestcase(yyruleno==109);
      case 110: /* value ::= FLOATING */ yytestcase(yyruleno==110);
      case 111: /* value ::= INTEGER */ yytestcase(yyruleno==111);
      case 112: /* value ::= DOTED_IP */ yytestcase(yyruleno==112);
#line 726 "mate_grammar.lemon"
{ yylhsminor.yy88 = g_strdup(yymsp[0].minor.yy0); }
#line 2365 "mate_grammar.c"
  yymsp[0].minor.yy88 = yylhsminor.yy88;
        break;
      case 113: /* value ::= COLONIZED */
#line 731 "mate_grammar.lemon"
{ yylhsminor.yy88 = recolonize(mc,yymsp[0].minor.yy0); }
#line 2371 "mate_grammar.c"
  yymsp[0].minor.yy88 = yylhsminor.yy88;
        break;
      case 123: /* decl ::= DONE_KW SEMICOLON */
{  yy_destructor(yypParser,1,&yymsp[-1].minor);
#line 270 "mate_grammar.lemon"
{
}
#line 2379 "mate_grammar.c"
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 124: /* debug_decl ::= DEBUG_KW OPEN_BRACE dbgfile_default dbglevel_default pdu_dbglevel_default gop_dbglevel_default gog_dbglevel_default CLOSE_BRACE SEMICOLON */
{  yy_destructor(yypParser,3,&yymsp[-8].minor);
#line 275 "mate_grammar.lemon"
{
}
#line 2388 "mate_grammar.c"
  yy_destructor(yypParser,4,&yymsp[-7].minor);
  yy_destructor(yypParser,5,&yymsp[-1].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 130: /* defaults_decl ::= DEFAULT_KW OPEN_BRACE pdu_defaults gop_defaults gog_defaults CLOSE_BRACE SEMICOLON */
{  yy_destructor(yypParser,14,&yymsp[-6].minor);
#line 297 "mate_grammar.lemon"
{
}
#line 2399 "mate_grammar.c"
  yy_destructor(yypParser,4,&yymsp[-5].minor);
  yy_destructor(yypParser,5,&yymsp[-1].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 131: /* pdu_defaults ::= PDU_KW OPEN_BRACE pdu_last_extracted_default pdu_drop_unassigned_default pdu_discard_default CLOSE_BRACE SEMICOLON */
{  yy_destructor(yypParser,11,&yymsp[-6].minor);
#line 299 "mate_grammar.lemon"
{
}
#line 2410 "mate_grammar.c"
  yy_destructor(yypParser,4,&yymsp[-5].minor);
  yy_destructor(yypParser,5,&yymsp[-1].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 136: /* gop_defaults ::= GOP_KW OPEN_BRACE gop_expiration_default gop_idle_timeout_default gop_lifetime_default gop_drop_unassigned_default gop_tree_mode_default gop_show_times_default CLOSE_BRACE SEMICOLON */
{  yy_destructor(yypParser,12,&yymsp[-9].minor);
#line 311 "mate_grammar.lemon"
{
}
#line 2421 "mate_grammar.c"
  yy_destructor(yypParser,4,&yymsp[-8].minor);
  yy_destructor(yypParser,5,&yymsp[-1].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      case 144: /* gog_defaults ::= GOG_KW OPEN_BRACE gog_expiration_default gop_tree_mode_default gog_goptree_default gog_show_times_default CLOSE_BRACE SEMICOLON */
{  yy_destructor(yypParser,13,&yymsp[-7].minor);
#line 332 "mate_grammar.lemon"
{
}
#line 2432 "mate_grammar.c"
  yy_destructor(yypParser,4,&yymsp[-6].minor);
  yy_destructor(yypParser,5,&yymsp[-1].minor);
  yy_destructor(yypParser,2,&yymsp[0].minor);
}
        break;
      default:
      /* (114) mate_config ::= decls */ yytestcase(yyruleno==114);
      /* (115) decls ::= decls decl */ yytestcase(yyruleno==115);
      /* (116) decls ::= (OPTIMIZED OUT) */ assert(yyruleno!=116);
      /* (117) decl ::= pdu_decl (OPTIMIZED OUT) */ assert(yyruleno!=117);
      /* (118) decl ::= gop_decl (OPTIMIZED OUT) */ assert(yyruleno!=118);
      /* (119) decl ::= gog_decl (OPTIMIZED OUT) */ assert(yyruleno!=119);
      /* (120) decl ::= transform_decl (OPTIMIZED OUT) */ assert(yyruleno!=120);
      /* (121) decl ::= defaults_decl (OPTIMIZED OUT) */ assert(yyruleno!=121);
      /* (122) decl ::= debug_decl (OPTIMIZED OUT) */ assert(yyruleno!=122);
      /* (125) dbgfile_default ::= */ yytestcase(yyruleno==125);
      /* (126) dbglevel_default ::= */ yytestcase(yyruleno==126);
      /* (127) pdu_dbglevel_default ::= */ yytestcase(yyruleno==127);
      /* (128) gop_dbglevel_default ::= */ yytestcase(yyruleno==128);
      /* (129) gog_dbglevel_default ::= */ yytestcase(yyruleno==129);
      /* (132) pdu_defaults ::= */ yytestcase(yyruleno==132);
      /* (133) pdu_last_extracted_default ::= */ yytestcase(yyruleno==133);
      /* (134) pdu_drop_unassigned_default ::= */ yytestcase(yyruleno==134);
      /* (135) pdu_discard_default ::= */ yytestcase(yyruleno==135);
      /* (137) gop_defaults ::= */ yytestcase(yyruleno==137);
      /* (138) gop_expiration_default ::= */ yytestcase(yyruleno==138);
      /* (139) gop_idle_timeout_default ::= */ yytestcase(yyruleno==139);
      /* (140) gop_lifetime_default ::= */ yytestcase(yyruleno==140);
      /* (141) gop_drop_unassigned_default ::= */ yytestcase(yyruleno==141);
      /* (142) gop_tree_mode_default ::= */ yytestcase(yyruleno==142);
      /* (143) gop_show_times_default ::= */ yytestcase(yyruleno==143);
      /* (145) gog_defaults ::= */ yytestcase(yyruleno==145);
      /* (146) gog_expiration_default ::= */ yytestcase(yyruleno==146);
      /* (147) gog_goptree_default ::= */ yytestcase(yyruleno==147);
      /* (148) gog_show_times_default ::= */ yytestcase(yyruleno==148);
        break;
/********** End reduce actions ************************************************/
  };
  assert( yyruleno<sizeof(yyRuleInfo)/sizeof(yyRuleInfo[0]) );
  yygoto = yyRuleInfo[yyruleno].lhs;
  yysize = yyRuleInfo[yyruleno].nrhs;
  yyact = yy_find_reduce_action(yymsp[-yysize].stateno,(YYCODETYPE)yygoto);
  if( yyact <= YY_MAX_SHIFTREDUCE ){
    if( yyact>YY_MAX_SHIFT ){
      yyact += YY_MIN_REDUCE - YY_MIN_SHIFTREDUCE;
    }
    yymsp -= yysize-1;
    yypParser->yytos = yymsp;
    yymsp->stateno = (YYACTIONTYPE)yyact;
    yymsp->major = (YYCODETYPE)yygoto;
    yyTraceShift(yypParser, yyact);
  }else{
    assert( yyact == YY_ACCEPT_ACTION );
    yypParser->yytos -= yysize;
    yy_accept(yypParser);
  }
}

/*
** The following code executes when the parse fails
*/
#ifndef YYNOERRORRECOVERY
static void yy_parse_failed(
  yyParser *yypParser           /* The parser */
){
  MateParserARG_FETCH;
#ifndef NDEBUG
  if( yyTraceFILE ){
    fprintf(yyTraceFILE,"%sFail!\n",yyTracePrompt);
  }
#endif
  while( yypParser->yytos>yypParser->yystack ) yy_pop_parser_stack(yypParser);
  /* Here code is inserted which will be executed whenever the
  ** parser fails */
/************ Begin %parse_failure code ***************************************/
#line 199 "./mate_grammar.lemon"

	configuration_error(mc,"Parse Error");
#line 2511 "mate_grammar.c"
/************ End %parse_failure code *****************************************/
  MateParserARG_STORE; /* Suppress warning about unused %extra_argument variable */
}
#endif /* YYNOERRORRECOVERY */

/*
** The following code executes when a syntax error first occurs.
*/
static void yy_syntax_error(
  yyParser *yypParser,           /* The parser */
  int yymajor _U_,               /* The major type of the error token */
  MateParserTOKENTYPE yyminor         /* The minor type of the error token */
){
  MateParserARG_FETCH;
#define TOKEN yyminor
/************ Begin %syntax_error code ****************************************/
#line 195 "./mate_grammar.lemon"

	configuration_error(mc,"Syntax Error before %s",yyminor);
#line 2531 "mate_grammar.c"
/************ End %syntax_error code ******************************************/
  MateParserARG_STORE; /* Suppress warning about unused %extra_argument variable */
}

/*
** The following is executed when the parser accepts
*/
static void yy_accept(
  yyParser *yypParser           /* The parser */
){
  MateParserARG_FETCH;
#ifndef NDEBUG
  if( yyTraceFILE ){
    fprintf(yyTraceFILE,"%sAccept!\n",yyTracePrompt);
  }
#endif
#ifndef YYNOERRORRECOVERY
  yypParser->yyerrcnt = -1;
#endif
  assert( yypParser->yytos==yypParser->yystack );
  /* Here code is inserted which will be executed whenever the
  ** parser accepts */
/*********** Begin %parse_accept code *****************************************/
/*********** End %parse_accept code *******************************************/
  MateParserARG_STORE; /* Suppress warning about unused %extra_argument variable */
}

/* The main parser program.
** The first argument is a pointer to a structure obtained from
** "MateParserAlloc" which describes the current state of the parser.
** The second argument is the major token number.  The third is
** the minor token.  The fourth optional argument is whatever the
** user wants (and specified in the grammar) and is available for
** use by the action routines.
**
** Inputs:
** <ul>
** <li> A pointer to the parser (an opaque structure.)
** <li> The major token number.
** <li> The minor token number.
** <li> An option argument of a grammar-specified type.
** </ul>
**
** Outputs:
** None.
*/
void MateParser(
  void *yyp,                   /* The parser */
  int yymajor,                 /* The major token code number */
  MateParserTOKENTYPE yyminor       /* The value for the token */
  MateParserARG_PDECL               /* Optional %extra_argument parameter */
){
  YYMINORTYPE yyminorunion;
  unsigned int yyact;   /* The parser action. */
#if !defined(YYERRORSYMBOL) && !defined(YYNOERRORRECOVERY)
  int yyendofinput;     /* True if we are at the end of input */
#endif
#ifdef YYERRORSYMBOL
  int yyerrorhit = 0;   /* True if yymajor has invoked an error */
#endif
  yyParser *yypParser;  /* The parser */

  yypParser = (yyParser*)yyp;
  assert( yypParser->yytos!=0 );
#if !defined(YYERRORSYMBOL) && !defined(YYNOERRORRECOVERY)
  yyendofinput = (yymajor==0);
#endif
  MateParserARG_STORE;

#ifndef NDEBUG
  if( yyTraceFILE ){
    fprintf(yyTraceFILE,"%sInput '%s'\n",yyTracePrompt,yyTokenName[yymajor]);
  }
#endif

  do{
    yyact = yy_find_shift_action(yypParser,(YYCODETYPE)yymajor);
    if( yyact <= YY_MAX_SHIFTREDUCE ){
      yy_shift(yypParser,yyact,yymajor,yyminor);
#ifndef YYNOERRORRECOVERY
      yypParser->yyerrcnt--;
#endif
      yymajor = YYNOCODE;
    }else if( yyact <= YY_MAX_REDUCE ){
      yy_reduce(yypParser,yyact-YY_MIN_REDUCE);
    }else{
      assert( yyact == YY_ERROR_ACTION );
      yyminorunion.yy0 = yyminor;
#ifdef YYERRORSYMBOL
      int yymx;
#endif
#ifndef NDEBUG
      if( yyTraceFILE ){
        fprintf(yyTraceFILE,"%sSyntax Error!\n",yyTracePrompt);
      }
#endif
#ifdef YYERRORSYMBOL
      /* A syntax error has occurred.
      ** The response to an error depends upon whether or not the
      ** grammar defines an error token "ERROR".
      **
      ** This is what we do if the grammar does define ERROR:
      **
      **  * Call the %syntax_error function.
      **
      **  * Begin popping the stack until we enter a state where
      **    it is legal to shift the error symbol, then shift
      **    the error symbol.
      **
      **  * Set the error count to three.
      **
      **  * Begin accepting and shifting new tokens.  No new error
      **    processing will occur until three tokens have been
      **    shifted successfully.
      **
      */
      if( yypParser->yyerrcnt<0 ){
        yy_syntax_error(yypParser,yymajor,yyminor);
      }
      yymx = yypParser->yytos->major;
      if( yymx==YYERRORSYMBOL || yyerrorhit ){
#ifndef NDEBUG
        if( yyTraceFILE ){
          fprintf(yyTraceFILE,"%sDiscard input token %s\n",
             yyTracePrompt,yyTokenName[yymajor]);
        }
#endif
        yy_destructor(yypParser, (YYCODETYPE)yymajor, &yyminorunion);
        yymajor = YYNOCODE;
      }else{
        while( yypParser->yytos >= yypParser->yystack
            && yymx != YYERRORSYMBOL
            && (yyact = yy_find_reduce_action(
                        yypParser->yytos->stateno,
                        YYERRORSYMBOL)) >= YY_MIN_REDUCE
        ){
          yy_pop_parser_stack(yypParser);
        }
        if( yypParser->yytos < yypParser->yystack || yymajor==0 ){
          yy_destructor(yypParser,(YYCODETYPE)yymajor,&yyminorunion);
          yy_parse_failed(yypParser);
#ifndef YYNOERRORRECOVERY
          yypParser->yyerrcnt = -1;
#endif
          yymajor = YYNOCODE;
        }else if( yymx!=YYERRORSYMBOL ){
          yy_shift(yypParser,yyact,YYERRORSYMBOL,yyminor);
        }
      }
      yypParser->yyerrcnt = 3;
      yyerrorhit = 1;
#elif defined(YYNOERRORRECOVERY)
      /* If the YYNOERRORRECOVERY macro is defined, then do not attempt to
      ** do any kind of error recovery.  Instead, simply invoke the syntax
      ** error routine and continue going as if nothing had happened.
      **
      ** Applications can set this macro (for example inside %include) if
      ** they intend to abandon the parse upon the first syntax error seen.
      */
      yy_syntax_error(yypParser,yymajor, yyminor);
      yy_destructor(yypParser,(YYCODETYPE)yymajor,&yyminorunion);
      yymajor = YYNOCODE;

#else  /* YYERRORSYMBOL is not defined */
      /* This is what we do if the grammar does not define ERROR:
      **
      **  * Report an error message, and throw away the input token.
      **
      **  * If the input token is $, then fail the parse.
      **
      ** As before, subsequent error messages are suppressed until
      ** three input tokens have been successfully shifted.
      */
      if( yypParser->yyerrcnt<=0 ){
        yy_syntax_error(yypParser,yymajor, yyminor);
      }
      yypParser->yyerrcnt = 3;
      yy_destructor(yypParser,(YYCODETYPE)yymajor,&yyminorunion);
      if( yyendofinput ){
        yy_parse_failed(yypParser);
#ifndef YYNOERRORRECOVERY
        yypParser->yyerrcnt = -1;
#endif
      }
      yymajor = YYNOCODE;
#endif
    }
  }while( yymajor!=YYNOCODE && yypParser->yytos>yypParser->yystack );
#ifndef NDEBUG
  if( yyTraceFILE ){
    yyStackEntry *i;
    char cDiv = '[';
    fprintf(yyTraceFILE,"%sReturn. Stack=",yyTracePrompt);
    for(i=&yypParser->yystack[1]; i<=yypParser->yytos; i++){
      fprintf(yyTraceFILE,"%c%s", cDiv, yyTokenName[i->major]);
      cDiv = ' ';
    }
    fprintf(yyTraceFILE,"]\n");
  }
#endif
  return;
}
